"""Configure explorer data restore definition."""
from concurrent.futures import ThreadPoolExecutor

from ..common import (
    GencoveExplorerManager,
    request_is_from_explorer,
    validate_explorer_user_data,
)
from ....base import Command
from ....utils import user_has_aws_in_path
from .....constants import RESTORE_TIERS_SUPPORTED
from .....exceptions import ValidationError


class Restore(Command):
    """Restore archived Explorer contents command executor."""

    def __init__(self, ctx, path, days, tier, credentials, options):
        super().__init__(credentials, options)
        self.ctx = ctx
        self.path = path
        self.days = days
        self.tier = tier

        # Populated after self.login() is called
        self.user = None
        self.organization = None
        self.aws_session_credentials = None

    def validate(self):
        """Validate restore"""
        self.validate_tier()
        validate_explorer_user_data(self.user, self.organization)
        user_has_aws_in_path(raise_exception=True)

    def validate_tier(self):
        """Validate that tier is supported.

        Raises:
            ValidationError: If tier is not supported.
        """
        if self.tier not in RESTORE_TIERS_SUPPORTED:
            raise ValidationError(
                f"Tier can only be one of: {RESTORE_TIERS_SUPPORTED}."
            )

    def initialize(self):
        """Initialize restore subcommand."""
        self.login()
        self.user = self.api_client.get_user_details()
        self.organization = self.api_client.get_organization_details()

        if not request_is_from_explorer():
            self.aws_session_credentials = (
                self.api_client.get_explorer_data_credentials()
            )

    def execute(self):
        """Make a request to restore Explorer objects."""
        self.echo_debug("List Explorer contents.")

        explorer_manager = GencoveExplorerManager(
            aws_session_credentials=self.aws_session_credentials,
            user_id=str(self.user.id),
            organization_id=str(self.organization.id),
        )

        bucket, _ = (
            explorer_manager.translate_path_to_s3_path(self.path)
            .lstrip("s3://")
            .split("/", 1)
        )
        s3_client = explorer_manager.thread_safe_client("s3")

        def restore_archived(obj) -> bool:
            """Restores archived file by calling S3 API.

            Args:
                obj (dict): Contains the "Key" of the object to
                    be restored.

            Returns:
                bool: False if the object is not archived,
                    otherwise returns True.
            """
            self.echo_debug(f"restore_archived: {obj}")

            if obj.get("StorageClass") not in [
                "GLACIER",
                "DEEP_ARCHIVE",
            ]:
                return False  # skip non-archived files

            # If already restored it extends the availability window
            s3_client.restore_object(
                Bucket=bucket,
                Key=obj["Key"],
                RestoreRequest={
                    "Days": self.days,
                    "GlacierJobParameters": {
                        "Tier": self.tier,
                    },
                },
            )

            self.echo_data(f"restore: {obj['Key']}")

            return True

        paginated_response = explorer_manager.list_s3_objects(self.path)
        obj_counts = {"skipped": 0, "restored": 0}
        with ThreadPoolExecutor() as executor:
            for response in paginated_response:
                for restored in executor.map(restore_archived, response["Contents"]):
                    if restored:
                        obj_counts["restored"] += 1
                    else:
                        obj_counts["skipped"] += 1

        if obj_counts["restored"]:
            self.echo_info(
                f"Restoring {obj_counts['restored']} objects in {self.path}."
                f" They will be available for {self.days} days."
            )

        if obj_counts["skipped"]:
            self.echo_info(
                f"Skipped {obj_counts['skipped']} objects that are not archived."
            )



"""Test project's batches create command."""
# pylint: disable=wrong-import-order, import-error
import io
import operator
import sys
import uuid

from click import echo
from click.testing import CliRunner

from gencove.client import APIClient, APIClientError  # noqa: I100
from gencove.command.projects.cli import create_project
from gencove.command.utils import sanitize_string
from gencove.models import Project
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.projects.vcr.filters import (
    filter_create_project_request,
    filter_create_project_response,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_create_project_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
            filter_create_project_response,
        ],
    }


def test_create_project__bad_pipeline_capability_id(credentials, mocker):
    """Test project creation failure when non-uuid string is used as
    pipeline capability id.
    """
    runner = CliRunner()
    mocked_create_project = mocker.patch.object(
        APIClient,
        "create_project",
    )
    res = runner.invoke(
        create_project,
        [
            "foo",
            "1111111",
            *credentials,
        ],
    )
    assert res.exit_code == 1
    mocked_create_project.assert_not_called()
    assert "Pipeline capability ID is not valid" in res.output


@pytest.mark.vcr
@assert_authorization
def test_create_project__no_pipeline_capability(credentials, mocker, recording):
    """Test project creation failure when there's no pipeline capability."""
    runner = CliRunner()
    if not recording:
        mocked_create_project = mocker.patch.object(
            APIClient,
            "create_project",
            side_effect=APIClientError(message="", status_code=400),
        )

    res = runner.invoke(
        create_project,
        [
            "foo",
            str(uuid.uuid4()),
            *credentials,
        ],
    )
    assert res.exit_code == 1
    if not recording:
        mocked_create_project.assert_called_once()
        assert "WARNING: There was an error creating a project." in res.output


@pytest.mark.vcr
@assert_authorization
def test_create_project__success(
    credentials,
    mocker,
    pipeline_capability_id,
    recording,
    vcr,
):
    """Test project creation success."""
    runner = CliRunner()
    if not recording:
        # Mock create_project only if using the cassettes, since we
        # mock the return value.
        create_project_response = get_vcr_response(
            "/api/v2/projects/", vcr, operator.contains
        )
        mocked_create_project = mocker.patch.object(
            APIClient,
            "create_project",
            return_value=Project(**create_project_response),
        )
    res = runner.invoke(
        create_project,
        [
            "Test create project",
            pipeline_capability_id,
            *credentials,
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_create_project.assert_called_once()

        output_line = io.BytesIO()
        sys.stdout = output_line
        project = Project(**create_project_response)
        echo(
            "\t".join(
                [
                    str(project.created),
                    str(project.id),
                    sanitize_string(project.name),
                    str(project.pipeline_capabilities),
                ]
            )
        )
        assert output_line.getvalue() == res.output.encode()



"""File like object that reads multiple files as if they are one file."""
import os


# pylint: disable=R0205
class MultiFileReader(object):
    """File-like object that reads multi-files as if they are one file."""

    def __init__(self, files):
        if isinstance(files, str):
            self._files = (files,)
        else:
            self._files = files

        self._file = None
        self._file_idx = 0
        self._map_sizes = {
            file_path: os.path.getsize(file_path) for file_path in self._files
        }
        self._read_files_offset = {file_path: 0 for file_path in self._files}

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def __iter__(self):  # pylint: disable=E0301
        return self

    def close(self):
        """Close file and clear state."""
        if self._file:
            self._file.close()
            self._file = None
            self._file_idx = 0

    def nextfile(self):
        """Get next file if there is one."""
        prev_file = self._file
        self._file = None
        if prev_file:
            prev_file.close()
            self._file_idx += 1

        try:
            self._file = open(  # pylint: disable=consider-using-with
                self._files[self._file_idx], "rb"
            )
        except IndexError:
            pass

    def filename(self):
        """Returns filename of the current file."""
        return self._file.name

    def fileno(self):
        """Returns fileno of the current file."""
        if self._file:
            try:
                return self._file.fileno()
            except ValueError:
                return -1
        else:
            return -1

    def get_size(self):
        """Returns combined size of the files."""
        return sum(self._map_sizes.values())

    def read(self, size=None):
        """Read a chunk of the file.

        If the size is not provided, will read all files at once.

        Args:
            size (int): number of bytes to read

        Returns:
            bytes: the chunk
        """
        if not size:
            size = self.get_size()

        buf = b""
        while size > 0:
            if not self._file and self._file_idx == len(self._files):
                break

            # pylint: disable=E0012,C0330
            if (
                not self._file
                or self._read_files_offset[self._file.name]
                == self._map_sizes[self._file.name]
            ):
                self.nextfile()
                continue

            unread = (
                self._map_sizes[self._file.name]
                - self._read_files_offset[self._file.name]
            )
            length = min(size, unread)
            self._file.seek(self._read_files_offset[self._file.name])
            buf += self._file.read(length)
            size -= length
            self._read_files_offset[self._file.name] += length

        return buf



"""Utilities for processing batches."""


def get_line(batch):
    """Build a list of relevant data to be printed.

    Args:
        batch (dict): an object from project batches response

    Returns:
        str: tab delimited list of relevant data to be printed
    """
    return "\t".join(
        [
            str(batch.id),
            batch.last_status.created.isoformat(),
            batch.last_status.status,
            batch.batch_type,
            batch.name,
        ]  # noqa
    )



"""List project's batch types subcommand."""

import backoff

# pylint: disable=wrong-import-order
from gencove import client  # noqa: I100
from gencove.command.base import Command

from .utils import get_line


class ListBatchTypes(Command):
    """List batch types command executor."""

    def __init__(self, project_id, credentials, options):
        super().__init__(credentials, options)
        self.project_id = project_id

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""

    def execute(self):
        self.echo_debug("Retrieving project's batch types:")

        try:
            for batch_types in self.get_paginated_batch_types():
                if not batch_types:
                    self.echo_debug("No matching batch types were found.")
                    return

                for batch_type in batch_types:
                    self.echo_data(get_line(batch_type))

        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning("There was an error listing project batch types.")
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            elif err.status_code == 404:
                self.echo_warning(f"Project {self.project_id} does not exist.")
            else:
                raise

    def get_paginated_batch_types(self):
        """Paginate over all batch types for the destination.

        Yields:
            paginated lists of batch types
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get batch types page")
            resp = self.get_batch_types(next_link)
            yield resp.results
            next_link = resp.meta.next
            more = next_link is not None

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=2,
        max_time=30,
    )
    def get_batch_types(self, next_link=None):
        """Get batch types page."""
        return self.api_client.get_project_batch_types(
            project_id=self.project_id, next_link=next_link
        )



"""Webhook validation related utilites."""
import hashlib
import hmac

import requests


# Copied from:
# https://docs.gencove.com/base/analysis/event-notifications#webhook-signatures
def calculate_signature(secret, timestamp, payload):
    """Calculates hmac signature with the given secret"""
    signature_message = f"{timestamp}.{payload}".encode("utf-8")
    return hmac.new(
        secret.encode("utf-8"), signature_message, hashlib.sha512
    ).hexdigest()


def is_valid_signature(secret, header, payload):
    """Test if provided signature is a valid.

    secret (str): key to be used as a secret for hmac algorithm.
    header (str): Gencove-Signature header content.
    payload (str): JSON payload (i.e., the request’s body).

    Returns:
        bool: True if is a valid signature, False if not
    """
    header = requests.utils.parse_dict_header(header)

    if "v1" not in header or "t" not in header:
        return False

    return calculate_signature(secret, header["t"], payload) == header["v1"]



"""Get sample manifest shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_destination_exists, validate_uuid
from gencove.constants import Credentials, Optionals

from .main import GetSampleManifest


@click.command("get-sample-manifest")
@click.argument("manifest_id", callback=validate_uuid)
@click.argument("destination", callback=validate_destination_exists)
@add_options(common_options)
def get_sample_manifest(  # pylint: disable=too-many-arguments
    manifest_id,
    destination,
    host,
    email,
    password,
    api_key,
):
    """Download a sample manifest to a specified destination directory.

    `MANIFEST_ID`: Gencove manifest ID

    `DESTINATION`: path/to/save/manifest/to
    """
    GetSampleManifest(
        manifest_id,
        destination,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Test data ls command."""
import io
import sys

# pylint: disable=wrong-import-order, import-error

from click import echo
from click.testing import CliRunner

from gencove.client import (
    APIClient,
    APIClientError,
)  # noqa: I100
from gencove.command.explorer.data.cli import ls
from gencove.command.explorer.data.common import GencoveExplorerManager
from gencove.models import AWSCredentials
from gencove.tests.decorators import assert_authorization
from gencove.tests.explorer.vcr.filters import (  # noqa: I101
    filter_data_credentials_response,
)
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/explorer/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
            filter_data_credentials_response,
        ],
    }


@pytest.mark.vcr
@assert_authorization
def test_data_ls_success(mocker, credentials, recording, vcr):
    """Test data being output to shell."""
    runner = CliRunner()
    mocker.patch(
        "gencove.command.explorer.data.ls.main.user_has_aws_in_path", return_value=True
    )
    if not recording:
        credentials_response = get_vcr_response(
            "/api/v2/explorer-data-credentials/", vcr
        )
        mocked_get_credentials = mocker.patch.object(
            APIClient,
            "get_explorer_data_credentials",
            return_value=AWSCredentials(**credentials_response),
        )
        mocked_aws = mocker.patch.object(
            GencoveExplorerManager, "execute_aws_s3_path", return_value=None
        )

    # Assumption that CLI user has a file called cli_test_file.txt in their
    # Explorer files dir
    res = runner.invoke(ls, ["e://users/me/", *credentials])

    if not recording:
        mocked_get_credentials.assert_called_once()
        mocked_aws.assert_called_once()
    else:
        assert "cli_test_file.txt" in res.output

    assert res.exit_code == 0


@pytest.mark.vcr
@assert_authorization
def test_data_ls_no_permission(mocker, credentials):
    """Test no permissions for credentials endpoint."""
    runner = CliRunner()
    mocker.patch(
        "gencove.command.explorer.data.ls.main.user_has_aws_in_path", return_value=True
    )
    mocked_get_credentials = mocker.patch.object(
        APIClient,
        "get_explorer_data_credentials",
        side_effect=APIClientError(
            message="API Client Error: Not Found: Not found.", status_code=403
        ),
        return_value={"detail": "Not found"},
    )

    res = runner.invoke(ls, ["e://users/me/", *credentials])

    assert res.exit_code == 1

    mocked_get_credentials.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\n".join(
            [
                "ERROR: You do not have the sufficient permission "
                "level required to perform this operation.",
                "ERROR: API Client Error: Not Found: Not found.",
                "Aborted!",
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()



"""Get project QC report shell command definition."""

import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_uuid
from gencove.constants import Credentials, Optionals

from .main import ProjectQCReport


@click.command("project-qc")
@click.argument("project_id", callback=validate_uuid)
@click.option(
    "--columns",
    default=None,
    help="A comma separated list of columns to include in the report. "
    "By default, all columns are included.",
)
@click.option(
    "--output-filename",
    help="Output filename for the output report. "
    "If not supplied, will download the file to the current directory.",
    type=click.Path(),
    default=None,
    required=False,
)
@add_options(common_options)
def project_qc(  # pylint: disable=too-many-arguments
    project_id,
    columns,
    output_filename,
    host,
    email,
    password,
    api_key,
):
    """Download a project QC CSV report to a specified filename.

    `PROJECT_ID`: Gencove project ID
    """
    ProjectQCReport(
        project_id,
        columns,
        output_filename,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Configure explorer data restore definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import Restore


@click.command(
    "restore",
    help="Restore archived data in Explorer object storage",
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument("path", type=click.Path(), default="e://")
@click.option(
    "--days", default=30, help="Days that the restored data will be available for."
)
@click.option(
    "--tier",
    default="Bulk",
    help=(
        "When restoring an archived object, "
        "you can specify one of the following data access tier options: "
        "Expedited, Standard, Bulk."
    ),
)
@click.pass_context
@add_options(common_options)
def restore(  # pylint: disable=too-many-arguments,invalid-name
    ctx,
    path,
    days,
    tier,
    host,
    email,
    password,
    api_key,
):  # pylint: disable=line-too-long
    """Restore archived objects."""  # noqa: E501
    Restore(
        ctx,
        path,
        days,
        tier,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Configure presign for explorer data command definition.
"""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import Presign


@click.command(
    "presign",
    help="Generate a presigned URL for an object in Explorer object storage",
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument("path", type=click.Path())
@click.pass_context
@add_options(common_options)
def presign(  # pylint: disable=too-many-arguments,invalid-name
    ctx,
    path,
    host,
    email,
    password,
    api_key,
):  # pylint: disable=line-too-long
    """Start instances."""  # noqa: E501
    Presign(
        ctx,
        path,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Util functions."""
import json
import operator

from requests import Response
from requests.structures import CaseInsensitiveDict

MOCK_UUID = "11111111-1111-1111-1111-111111111111"
MOCK_CHECKSUM = "111111111111111111111111111111111111111111111111111111111111111a"


def get_vcr_response(url, vcr, matches=operator.eq, just_body=True):
    """Returns the VCR response of the first request that matches the url.

    Args:
        url (str): The request URL you want to match
        vcr (Cassette): VCR object, use the fixture
        matches (callable): Evaluator function
        just_body (bool): Set to False to retrieve the entire response object.

    Returns:
        response (dict): The VCR response.
    """
    request = next(request for request in vcr.requests if matches(request.path, url))
    response = vcr.responses_of(request)[0]
    if just_body:
        response = response["body"]["string"]
        try:
            response = json.loads(response)
        except json.decoder.JSONDecodeError:
            pass
    return response


def get_response_from_vcr_dict(vcr_dict: dict) -> Response:
    """Create Response object from get_vcr_response return value"""
    response = Response()
    response.status_code = vcr_dict["status"]["code"]
    response.headers = CaseInsensitiveDict(vcr_dict["headers"])
    response._content = vcr_dict["body"]["string"]  # pylint: disable=protected-access
    return response



"""Commands to be executed from command line."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, FileTypesObject, Optionals
from gencove.utils import enum_as_dict

from .main import ListFileTypes


@click.command("file-types")
@click.option("--project-id", required=False, help="Gencove project ID", type=str)
@click.option(
    "--object",
    "object_param",
    required=False,
    help="File types of a specific object",
    type=click.Choice(enum_as_dict(FileTypesObject).values()),
    default=FileTypesObject.SAMPLE.value,
)
@add_options(common_options)
def list_file_types(project_id, object_param, host, email, password, api_key):
    """List file types in Gencove's system."""
    ListFileTypes(
        project_id,
        object_param,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Commands to be executed from command line."""
# pylint: disable=E0012,C0330,R0913
import click

from .create.cli import create_project
from .create_batch.cli import create_project_batch
from .create_merged_vcf.cli import create_merged_vcf
from .create_sample_manifest.cli import create_sample_manifest
from .delete.cli import delete_projects
from .delete_samples.cli import delete_project_samples
from .get_batch.cli import get_batch
from .get_merged_vcf.cli import get_merged_vcf
from .get_reference_genome.cli import get_reference_genome
from .get_sample_manifests.cli import get_sample_manifests
from .import_existing_samples.cli import import_existing_project_samples
from .list.cli import list_projects
from .list_batch_types.cli import list_project_batch_types
from .list_batches.cli import list_project_batches
from .list_pipeline_capabilities.cli import list_project_pipeline_capabilities
from .list_pipelines.cli import list_project_pipelines
from .restore_samples.cli import restore_project_samples
from .run_prefix.cli import run_prefix
from .samples.cli import list_project_samples
from .status_merged_vcf.cli import status_merged_vcf


@click.group()
def projects():
    """Project managements commands."""


projects.add_command(create_project)
projects.add_command(create_project_batch)
projects.add_command(list_projects)
projects.add_command(list_project_samples)
projects.add_command(list_project_batch_types)
projects.add_command(list_project_batches)
projects.add_command(list_project_pipeline_capabilities)
projects.add_command(list_project_pipelines)
projects.add_command(get_batch)
projects.add_command(delete_project_samples)
projects.add_command(restore_project_samples)
projects.add_command(import_existing_project_samples)
projects.add_command(run_prefix)
projects.add_command(create_merged_vcf)
projects.add_command(status_merged_vcf)
projects.add_command(get_merged_vcf)
projects.add_command(delete_projects)
projects.add_command(create_sample_manifest)
projects.add_command(get_sample_manifests)
projects.add_command(get_reference_genome)



"""Commands to be executed from command line."""
import shutil
import subprocess  # nosec B404 (bandit subprocess import)
import sys

import click

# This package enables Click to default to a command if it is not specified.
from click_default_group import DefaultGroup


from .data.cli import data
from .instances.cli import instances


def explorer_cli_installed():
    """Function that checks if the Explorer CLI is installed and accessible on
    the current system"""
    return shutil.which("explorer") is not None


class CustomEpilogDefaultGroup(DefaultGroup):
    """Click group class that inherits functionality from DefaultGroup and adds
    a custom epilog to collect top-level Explorer CLI command help"""

    def format_epilog(self, ctx, formatter):
        """Adds an epilog to describe Explorer CLI commands"""
        if explorer_cli_installed():
            try:
                result = subprocess.run(  # nosec B607 (start_process_with_partial_path)
                    ["explorer", "internal", "print-commands"],
                    capture_output=True,
                    text=True,
                    check=True,
                )
                formatter.write(result.stdout)
            except subprocess.CalledProcessError:
                # Ignore errors to fail gracefully if installed version of
                # explorer CLI does not support `print-commands`
                pass


@click.group(
    cls=CustomEpilogDefaultGroup,
    default_if_no_args=False,
    default="default",
)
def explorer():
    """Gencove Explorer commands."""


# # API-related Gencove CLI commands would be added like standard Click commands
# # in the CLI like in the example below
# @explorer.command()
# def test():
#     print("test")


# This method handles commands that have not been explicitly defined above
@explorer.command(
    hidden=True,
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
    add_help_option=False,
)
@click.pass_context
def default(ctx):
    """Default command to execute"""
    if not explorer_cli_installed():
        click.echo(
            "It doesn't seem like this command has been run from the Explorer "
            "ecosystem. Please try running it again from within Explorer.",
            err=True,
        )
        sys.exit(1)
    try:
        subprocess.run(  # nosec B603 (execution of untrusted input)
            ["explorer"] + ctx.args,
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
            check=True,
        )
    except subprocess.CalledProcessError as exception:
        sys.exit(exception.returncode)


explorer.add_command(instances)
explorer.add_command(data)



"""Get sample metadata subcommand."""

import json
import os

import backoff

from ...base import Command
from ...utils import is_valid_uuid
from .... import client
from ....exceptions import ValidationError


class GetMetadata(Command):
    """Get metadata command executor."""

    def __init__(self, sample_id, output_filename, credentials, options):
        super().__init__(credentials, options)
        self.sample_id = sample_id
        self.output_filename = output_filename

    def initialize(self):
        """Initialize get-metadata subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """

        if is_valid_uuid(self.sample_id) is False:
            raise ValidationError("Sample ID is not valid. Exiting.")

    def execute(self):
        self.echo_debug("Retrieving sample metadata:")

        try:
            metadata = self.get_metadata()
            self.output_metadata(metadata)

        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 404:
                self.echo_error(f"Sample metadata {self.sample_id} does not exist.")
            raise

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=2,
        max_time=30,
    )
    def get_metadata(self):
        """Get metadata page."""
        response = self.api_client.get_metadata(sample_id=self.sample_id)
        # API always returns a dictionary with a "metadata" key
        return response.metadata

    def output_metadata(self, metadata):
        """Output reformatted metadata JSON."""
        self.echo_debug("Outputting JSON.")
        if self.output_filename == "-":
            self.echo_data(json.dumps(metadata, indent=4, cls=client.CustomEncoder))
        else:
            dirname = os.path.dirname(self.output_filename)
            if dirname and not os.path.exists(dirname):
                os.makedirs(dirname)
            with open(self.output_filename, "w", encoding="utf-8") as json_file:
                json_file.write(
                    json.dumps(metadata, indent=4, cls=client.CustomEncoder)
                )
            self.echo_info(f"Sample metadata saved to {self.output_filename}")



"""Tests for loggers of Gencove CLI."""
import os

from click.testing import CliRunner

from gencove.logger import _log, dump_debug_log


def test_dump_debug_log(dump_filename):
    """Tests that the logs sent to _log are saved into a file when calling
    dump_debug_log.
    """
    del os.environ["GENCOVE_SAVE_DUMP_LOG"]
    logs = [f"log{i}" for i in range(100)]
    for msg in logs:
        _log(msg)
    runner = CliRunner()
    with runner.isolated_filesystem():
        dump_debug_log()
        with open(dump_filename, encoding="utf8") as log_file:
            log_content = log_file.read()
            assert all(log in log_content for log in logs)



"""Test data restore command."""
import io
import sys

# pylint: disable=wrong-import-order, import-error

from click import echo
from click.testing import CliRunner

from gencove.client import (
    APIClient,
    APIClientError,
)  # noqa: I100
from gencove.command.explorer.data.cli import restore
from gencove.command.explorer.data.common import GencoveExplorerManager
from gencove.models import AWSCredentials
from gencove.tests.decorators import assert_authorization
from gencove.tests.explorer.vcr.filters import (  # noqa: I101
    filter_data_credentials_response,
)
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/explorer/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
            filter_data_credentials_response,
        ],
    }


@pytest.mark.vcr
@assert_authorization
def test_data_restore_success(mocker, credentials, recording, vcr):
    """Test data being output to shell."""
    runner = CliRunner()
    mocker.patch(
        "gencove.command.explorer.data.restore.main.user_has_aws_in_path",
        return_value=True,
    )
    mocked_thread_safe_client = mocker.patch.object(
        GencoveExplorerManager, "thread_safe_client"
    )
    objects_in_storage = [
        {
            "Key": f"example{i}.fastq.gz",
            "StorageClass": "DEEP_ARCHIVE" if i % 2 == 0 else "STANDARD",
        }
        for i in range(100)
    ]
    mocked_list_objects = mocker.patch.object(
        GencoveExplorerManager,
        "list_s3_objects",
        return_value=[{"Contents": objects_in_storage}],
    )
    if not recording:
        credentials_response = get_vcr_response(
            "/api/v2/explorer-data-credentials/", vcr
        )
        mocked_get_credentials = mocker.patch.object(
            APIClient,
            "get_explorer_data_credentials",
            return_value=AWSCredentials(**credentials_response),
        )

    restore_path = "e://users/me/"
    res = runner.invoke(restore, [restore_path, *credentials])
    assert res.exit_code == 0

    if not recording:
        mocked_get_credentials.assert_called_once()
    mocked_thread_safe_client.assert_called_with("s3")
    mocked_list_objects.assert_called_with(restore_path)
    assert (
        mocked_thread_safe_client.return_value.restore_object.call_count
        == len(objects_in_storage) // 2
    )


@pytest.mark.vcr
@assert_authorization
def test_data_restore_no_permission(mocker, credentials):
    """Test no permissions for credentials endpoint."""
    runner = CliRunner()
    mocker.patch(
        "gencove.command.explorer.data.restore.main.user_has_aws_in_path",
        return_value=True,
    )
    mocked_get_credentials = mocker.patch.object(
        APIClient,
        "get_explorer_data_credentials",
        side_effect=APIClientError(
            message="API Client Error: Not Found: Not found.", status_code=403
        ),
        return_value={"detail": "Not found"},
    )

    res = runner.invoke(restore, ["e://users/me/", *credentials])

    assert res.exit_code == 1

    mocked_get_credentials.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\n".join(
            [
                "ERROR: You do not have the sufficient permission "
                "level required to perform this operation.",
                "ERROR: API Client Error: Not Found: Not found.",
                "Aborted!",
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()



"""Samples list shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_uuid
from gencove.constants import (
    Credentials,
    SampleArchiveStatus,
    SampleStatus,
)
from gencove.utils import enum_as_dict

from .constants import SamplesOptions
from .main import ListSamples


@click.command("list-samples")
@click.argument("project_id", callback=validate_uuid)
@click.option("--search", help="Gencove sample ID, client ID or metadata substring")
@click.option(
    "--status",
    help="Get samples with specific status",
    type=click.Choice(enum_as_dict(SampleStatus).values()),
    default=SampleStatus.ALL.value,
)
@click.option(
    "--archive-status",
    help="Get samples with specific archive status",
    type=click.Choice(enum_as_dict(SampleArchiveStatus).values()),
    default=SampleArchiveStatus.ALL.value,
)
@click.option(
    "--include-run/--no-include-run",
    help="Include run id",
    default=False,
)
@add_options(common_options)
def list_project_samples(  # pylint: disable=E0012,C0330,R0913
    project_id,
    search,
    status,
    archive_status,
    include_run,
    host,
    email,
    password,
    api_key,
):
    """List samples in a project.

    `PROJECT_ID`: Gencove project ID
    """
    ListSamples(
        project_id,
        Credentials(email=email, password=password, api_key=api_key),
        SamplesOptions(
            host=host,
            status=status,
            archive_status=archive_status,
            search=search,
            include_run=include_run,
        ),
    ).run()



"""Utilities for processing pipeline capabilities."""


def get_line(pipeline_capability):
    """Build a list of relevant data to be printed.

    Args:
        pipeline_capability (dict): an object from pipeline capabilities response

    Returns:
        str: tab delimited list of relevant data to be printed
    """
    return "\t".join(
        [
            str(pipeline_capability.id),
            pipeline_capability.name,
        ]
    )



"""List projects subcommand."""
import backoff

# pylint: disable=wrong-import-order
from gencove.client import APIClientError, APIClientTimeout  # noqa: I100
from gencove.command.base import Command
from gencove.models import Project

from .utils import get_line


class List(Command):
    """List projects command executor."""

    def __init__(self, include_capability, credentials, options):
        super().__init__(credentials, options)
        self.include_capability = include_capability

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""

    def execute(self):
        self.echo_debug("Retrieving projects")

        try:
            for projects in self.get_paginated_projects():
                if not projects:
                    self.echo_debug("No projects were found.")
                    return

                augmented_projects = (
                    self.augment_projects_with_pipeline_capabilities(  # noqa: E501
                        projects
                    )
                )

                for project in augmented_projects:
                    self.echo_data(get_line(project, self.include_capability))
        except APIClientError as err:
            if err.status_code == 404:
                self.echo_error("No projects found.")
            raise

    def get_paginated_projects(self):
        """Paginate over all projects.

        Yields:
            paginated lists of projects
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get projects page")
            resp = self.get_projects(next_link)

            yield resp.results
            next_link = resp.meta.next
            more = next_link is not None

    @backoff.on_exception(
        backoff.expo,
        (APIClientTimeout),
        max_tries=2,
        max_time=30,
    )
    def get_projects(self, next_link=None):
        """Get projects page."""
        return self.api_client.list_projects(next_link)

    @backoff.on_exception(
        backoff.expo,
        (APIClientTimeout),
        max_tries=3,
        max_time=30,
    )
    def get_pipeline_capabilities(self, pipeline_id):
        """Get pipeline capabilities."""
        return self.api_client.get_pipeline_capabilities(pipeline_id)

    def augment_projects_with_pipeline_capabilities(self, projects):
        """Fetch pipeline capabilities and append it to the project.

        Args:
            projects (list[Project]): list of projects

        Returns:
            list[Project]: same list of projects with pipeline capabilities
                uuid replaced with PipelineCapability
        """
        for project in projects:
            pipeline_capabilities = self.get_pipeline_capabilities(
                project.pipeline_capabilities
            )
            project_dict = dict(project)
            project_dict["pipeline_capabilities"] = pipeline_capabilities
            yield Project(**project_dict)



"""Download command executor."""
import json
import re
from pathlib import Path

import backoff

import requests

from gencove import client  # noqa: I100
from gencove.command.base import Command
from gencove.command.download.exceptions import DownloadTemplateError
from gencove.command.utils import validate_file_types
from gencove.constants import SampleArchiveStatus
from gencove.exceptions import ValidationError

from .constants import (
    ALLOWED_ARCHIVE_STATUSES_RE,
    ALLOWED_STATUSES_RE,
    METADATA_FILE_TYPE,
    QC_FILE_TYPE,
)
from .utils import (
    build_file_path,
    download_file,
    fatal_process_sample_error,
    get_download_template_format_params,
    save_metadata_file,
    save_qc_file,
)


# pylint: disable=too-many-instance-attributes
class Download(Command):
    """Download command executor."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        download_to,
        filters,
        credentials,
        options,
        download_urls,
        no_progress,
        checksums,
    ):
        super().__init__(credentials, options)
        self.download_to = download_to
        self.filters = filters
        self.options = options
        self.sample_ids = set()
        self.archived_samples_count = 0
        self.downloaded_files = set()
        self.download_urls = download_urls
        self.download_files = []
        self.no_progress = no_progress
        self.checksums = checksums

    def initialize(self):
        """Initialize download command."""
        if self.filters.project_id and self.filters.sample_ids:
            raise ValidationError("Must specify only one of: project-id or sample-ids")

        self.login()

        if self.filters.project_id:
            self.echo_debug(
                f"Retrieving sample ids for a project: {self.filters.project_id}"
            )

            try:
                samples_generator = self._get_paginated_samples()
                for sample in samples_generator:
                    if ALLOWED_ARCHIVE_STATUSES_RE.match(
                        sample.archive_last_status.status
                    ):
                        self.sample_ids.add(sample.id)
                    else:
                        self.archived_samples_count += 1
            except client.APIClientError as err:
                raise ValidationError(
                    f"Project id {self.filters.project_id} not found."
                ) from err
        else:
            self.sample_ids = self.filters.sample_ids

    def validate(self):
        """Validate command configuration before execution.

        Raises:
            ValidationError : something is wrong with configuration
        """
        if not self.filters.project_id and not self.filters.sample_ids:
            raise ValidationError("Must specify one of: project-id or sample-ids")

        if self.filters.project_id and self.filters.sample_ids:
            raise ValidationError("Must specify only one of: project-id or sample-ids")

        if not self.sample_ids:
            if self.archived_samples_count:
                raise ValidationError(
                    "No available samples to process. "
                    "All samples are in archive; "
                    "use the `gencove projects restore-samples` command to "
                    "restore before downloading. "
                    "Exiting."
                )
            raise ValidationError("No available samples to process. Exiting.")

        try:
            if self.filters.project_id:
                valid_file_types = self.api_client.get_file_types(
                    project_id=self.filters.project_id
                ).results
            else:
                valid_file_types = self.api_client.get_file_types().results
            invalid_file_types = validate_file_types(
                self.filters.file_types, valid_file_types
            )

            if invalid_file_types:
                raise ValidationError(
                    f"Invalid file types: {', '.join(invalid_file_types)}. "
                    f"Use gencove file-types command for a list of valid file types. "
                    f"Use with --project-id option to see project file types."
                )
        except client.APIClientError as err:
            self.echo_debug(err)
            self.echo_warning("There was an error while validating file types.")

        if all(
            [
                self.download_to == "-",
                not self.download_urls,
            ]
        ):
            raise ValidationError(
                "Cannot have - as a destination without download-urls."
            )

        if self.download_urls:
            self.echo_debug("Requesting download list in JSON format.")
        if self.download_to == "-":
            self.echo_debug("Will not download. Redirecting to STDOUT.")
        else:
            self.echo_debug(f"Host is {self.options.host} saving to {self.download_to}")

    def execute(self):
        if self.download_to != "-":
            self.echo_info("Processing samples")
        for sample_id in self.sample_ids:
            self.process_sample(sample_id)
        if self.download_urls:
            self.output_list()

        if all(
            [self.download_to != "-", not self.download_urls, not self.downloaded_files]
        ):
            sample_ids = [str(sample_id) for sample_id in self.sample_ids]
            self.echo_warning(
                f"Files not found for sample ids: {', '.join(sample_ids)} "
                f"and file types: {', '.join(self.filters.file_types)}."
            )
        else:
            self.echo_info(f"Processed {len(self.sample_ids)} samples")

        if self.archived_samples_count:
            self.echo_warning(
                f"Skipped {self.archived_samples_count} archived samples. "
                "Use the `gencove projects restore-samples` command to "
                "restore before downloading."
            )

    @backoff.on_exception(
        backoff.expo,
        requests.exceptions.HTTPError,
        giveup=fatal_process_sample_error,
        max_tries=10,
    )
    @backoff.on_exception(
        backoff.expo,
        client.APIClientError,
        max_tries=3,
    )
    @backoff.on_exception(
        backoff.expo,
        client.APIClientTooManyRequestsError,
        max_tries=20,
    )
    @backoff.on_exception(
        backoff.expo,
        client.APIClientTimeout,
        max_tries=20,
    )
    def process_sample(self, sample_id):
        """Process sample.

        Check if a sample is in appropriate state and if it is,
        get its files one by one.

        If downloading and a download failed with error 403, reprocess the
        sample in order to get fresh download url.
        """
        try:
            sample = self.api_client.get_sample_details(sample_id)
        except client.APIClientTooManyRequestsError:
            self.echo_debug(
                f"Request was throttled for sample {sample_id} "
                "because of too many requests, trying again"
            )
            raise
        except client.APIClientTimeout:
            self.echo_debug(
                f"Request was throttled for sample {sample_id} "
                "because of timeout, trying again"
            )
            raise
        except client.APIClientError as err:
            self.echo_debug(
                f"Sample with id {sample_id} not accessible due to "
                f"API error, trying again. Error: {err}"
            )
            raise

        self.echo_debug(
            f"Processing sample id {sample.id}, status {sample.last_status.status}"
        )

        if not ALLOWED_ARCHIVE_STATUSES_RE.match(sample.archive_last_status.status):
            raise ValidationError(
                f"Sample with id {sample.id} is archived and cannot be downloaded - "
                "please restore the sample and try again."
            )

        if not ALLOWED_STATUSES_RE.match(sample.last_status.status):
            self.echo_warning(f"Sample #{sample.id} has no deliverables.")
            return

        file_types_re = re.compile("|".join(self.filters.file_types), re.IGNORECASE)

        file_with_prefix = self.options.download_template.format(
            **get_download_template_format_params(sample.client_id, sample.id)
        )
        self.echo_debug(f"file path with prefix is: {file_with_prefix}")
        if all(
            [
                self.download_to != "-",
                not self.download_urls,
                file_types_re.match(QC_FILE_TYPE),
            ]
        ):
            self.download_sample_qc_metrics(file_with_prefix, sample_id)

        if all(
            [
                self.download_to != "-",
                not self.download_urls,
                file_types_re.match(METADATA_FILE_TYPE),
            ]
        ):
            self.download_sample_metadata(file_with_prefix, sample_id)

        self.download_files.append(
            {
                "gencove_id": sample.id,
                "client_id": sample.client_id,
                "last_status": {
                    "id": sample.last_status.id,
                    "status": sample.last_status.status,
                    "created": sample.last_status.created,
                },
                "archive_last_status": {
                    "id": sample.archive_last_status.id,
                    "status": sample.archive_last_status.status,
                    "created": sample.archive_last_status.created,
                    "transition_cutoff": (sample.archive_last_status.transition_cutoff),
                },
                "files": {},
            }
        )

        for sample_file in sample.files:
            # pylint: disable=E0012,C0330
            if self.filters.file_types and not file_types_re.match(
                sample_file.file_type
            ):
                self.echo_debug("Deliverable file type is not in desired file types")
                continue

            if not self.download_urls:
                file_path = build_file_path(
                    sample_file, file_with_prefix, self.download_to
                )
                self.validate_and_download(
                    file_path,
                    download_file,
                    file_path,
                    sample_file.download_url,
                    self.options.skip_existing,
                    self.no_progress,
                )
                if self.checksums:
                    try:
                        checksum = self.api_client.get_file_checksum(
                            sample_file.id, filename=Path(file_path).name
                        )
                        self.create_checksum_file(file_path, checksum)
                    except client.APIClientTooManyRequestsError:
                        self.echo_debug(
                            f"Request was throttled for file {sample_file}, "
                            f"trying again"
                        )
                        raise
            self.download_files[-1]["files"][sample_file.file_type] = {
                "id": sample_file.id,
                "download_url": sample_file.download_url,
                "checksum_sha256": sample_file.checksum_sha256,
            }

    def create_checksum_file(self, file_path, checksum_sha256):
        """Create checksum file.

        Args:
            file_path (str): File path of the original file,
                will append .sha256
            checksum_sha256 (str): Checksum (sha256) value for the file

        Returns:
            None
        """
        checksum_path = f"{file_path}.sha256"
        self.echo_debug(f"Adding checksum file: {checksum_path}")
        with open(checksum_path, "w", encoding="utf-8") as checksum_file:
            checksum_file.write(checksum_sha256)

    def validate_and_download(self, download_to_path, download_func, *args, **kwargs):
        """Check if this file was already downloaded, if yes - exit.

        Args:
            download_to_path(str): system file path to download to
            download_func(function): function that will do the download logic
            *args: arguments that will be passed to download function
            **kwargs: keyword arguments that will be passed to download func

        Returns:
            None

        Raises:
            DownloadTemplateError: if the file was found in already downloaded
             list
        """
        if download_to_path in self.downloaded_files:
            raise DownloadTemplateError(
                f"Bad template: {download_to_path} file already exists. "
                "Update your template to avoid files containing the same name "
                "and try again."
            )

        download_func(*args, **kwargs)

        self.echo_debug(f"Adding file path: {download_to_path}")
        self.downloaded_files.add(download_to_path)

    def download_sample_qc_metrics(self, file_with_prefix, sample_id):
        """Download and save to file on user file system.

        Args:
            file_with_prefix(str): file path based on download template,
                prefilled with sample data
            sample_id(str of uuid): sample gencove id

        Returns:
            None
        """
        file_path = build_file_path(
            dict(file_type=QC_FILE_TYPE),
            file_with_prefix,
            self.download_to,
            f"{sample_id}_{QC_FILE_TYPE}.json",
        )

        self.validate_and_download(
            file_path,
            save_qc_file,
            file_path,
            self.api_client,
            sample_id,
            self.options.skip_existing,
        )

    def download_sample_metadata(self, file_with_prefix, sample_id):
        """Get metadata and save to file on user file system.

        Args:
            file_with_prefix(str): file path based on download template,
                prefilled with sample data
            sample_id(str of uuid): sample gencove id

        Returns:
            None
        """
        file_path = build_file_path(
            dict(file_type=METADATA_FILE_TYPE),
            file_with_prefix,
            self.download_to,
            f"{sample_id}_{METADATA_FILE_TYPE}.json",
        )

        self.validate_and_download(
            file_path,
            save_metadata_file,
            file_path,
            self.api_client,
            sample_id,
            self.options.skip_existing,
        )

    def _get_paginated_samples(self):
        """Generate for project samples that traverses all pages."""
        get_samples = True
        next_page = None
        while get_samples:
            self.echo_debug(f"Getting page: {next_page or 1}")
            req = self.api_client.get_project_samples(
                self.filters.project_id,
                next_page,
                sample_archive_status=SampleArchiveStatus.ALL.value,
            )
            for sample in req.results:
                yield sample
            next_page = req.meta.next
            get_samples = next_page is not None

    def output_list(self):
        """Output reformatted JSON of each individual sample."""
        self.echo_debug("Outputting JSON.")
        if self.download_to == "-":
            self.echo_data(
                json.dumps(self.download_files, indent=4, cls=client.CustomEncoder)
            )
        else:
            with open(self.download_to, "w", encoding="utf-8") as json_file:
                json_file.write(
                    json.dumps(
                        self.download_files,
                        indent=4,
                        cls=client.CustomEncoder,
                    )
                )
            self.echo_info(
                "Samples and their deliverables download URLs outputted to "
                f"{self.download_to}"
            )



"""Configure inactivity stop for explorer instances subcommand."""


from ....base import Command
from .....exceptions import ValidationError


class StopInstances(Command):
    """Stop explorer instances command executor."""

    def validate(self):
        """Validate stop instances"""

    def initialize(self):
        """Initialize inactivity-stop subcommand."""
        self.login()

    def execute(self):
        """Make a request to stop explorer instances."""
        self.echo_debug("Stop explorer instances.")

        explorer_instances = self.api_client.get_explorer_instances()
        self.echo_debug(f"Found {len(explorer_instances.results)} explorer instances.")

        if len(explorer_instances.results) != 1:
            raise ValidationError(
                "Command not supported. Download the latest version of the Gencove CLI."
            )

        self.api_client.stop_explorer_instances(
            instance_ids=[e.id for e in explorer_instances.results]
        )
        self.echo_debug(
            f"Requested to stop {len(explorer_instances.results)} explorer instances."
        )

        self.echo_info("Request to stop explorer instances accepted.")



"""Create project executor."""
from gencove import client  # noqa: I100
from gencove.command.base import Command

from .utils import get_line


class CreateProject(Command):
    """Create project executor."""

    def __init__(
        self,
        project_name,
        pipeline_capability_id,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.project_name = project_name
        self.pipeline_capability_id = pipeline_capability_id

    def initialize(self):
        """Initialize create subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""

    def execute(self):
        """Make a request to create a project."""
        self.echo_debug(f"Creating a project {self.project_name}.")

        try:
            created_project_details = self.api_client.create_project(
                project_name=self.project_name,
                pipeline_capability_id=self.pipeline_capability_id,
            )
            self.echo_debug(created_project_details)
            self.echo_data(get_line(created_project_details))
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning("There was an error creating a project.")
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            raise



"""Utilities for processing Explorer instances."""

from .....models import ExplorerInstance


def get_line(explorer_instance: ExplorerInstance):
    """Build a list of relevant data to be printed.

    Args:
        explorer_instance (ExplorerInstance): instance of
            BaseSpace autoimport

    Returns:
        list(str): list of relevant data to be printed
    """
    return "\t".join(
        [
            str(explorer_instance.id),
            str(explorer_instance.status),
        ]
    )



"""Test webhook verify command."""

import hmac
import json
from datetime import datetime
from hashlib import sha512
from uuid import uuid4

from click.testing import CliRunner

from gencove.command.webhook.cli import verify


# Copied from back_api2/core/crypto.py#get_hmac
def get_hmac(secret_key, payload=None, digestmode=sha512, timestamp=None):
    """Generates hmac signature."""
    if payload is None:
        payload = ""

    if not timestamp:
        timestamp = int(datetime.utcnow().timestamp())

    return (
        timestamp,
        hmac.new(
            secret_key.encode("utf-8"),
            f"{timestamp}.{payload}".encode("utf-8"),
            digestmode,
        ).hexdigest(),
    )


INVALID_SECRET = "INVALID_SECRET"
VALID_SECRET = "VALID_SECRET"
VALID_PAYLOAD = json.dumps(
    [
        {
            "event_id": str(uuid4()),
            "event_type": "analysis_complete_v2",
            "timestamp": str(datetime.utcnow()),
            "payload": {
                "project": {"id": str(uuid4())},
                "samples": [
                    {
                        "id": str(uuid4()),
                        "client_id": str(uuid4()),
                        "last_status": {"status": "success"},
                    }
                ],
            },
        }
    ]
)
VALID_TIMESTAMP, VALID_SIGNATURE = get_hmac(VALID_SECRET, VALID_PAYLOAD)
VALID_HEADER = f"t={VALID_TIMESTAMP},v1={VALID_SIGNATURE}"

INVALID_PAYLOAD = "INVALID_PAYLOAD"
INVALID_HEADER_VERSION = f"t={VALID_TIMESTAMP},v2={VALID_SIGNATURE}"

INVALID_HEADER_FIELDS = f"t={VALID_TIMESTAMP}"
INVALID_HEADER_DATA = "t,v"
INVALID_HEADER_TIMESTAMP = f"t=INVALID_TIMESTAMP,v1={VALID_SIGNATURE}"
INVALID_HEADER_SIGNATURE = f"t={VALID_TIMESTAMP},v1=INVALID_SIGNATURE"


def test_verify__invalid_signature():
    """Test verify for all possible inputs of invalid signatures."""

    invalid_inputs = [
        # Test header version different from v1
        [
            VALID_SECRET,
            INVALID_HEADER_VERSION,
            VALID_PAYLOAD,
        ],
        # Test header invalid (missing fields)
        [
            VALID_SECRET,
            INVALID_HEADER_FIELDS,
            VALID_PAYLOAD,
        ],
        # Test header data (missing '=')
        [
            VALID_SECRET,
            INVALID_HEADER_DATA,
            VALID_PAYLOAD,
        ],
        # Test header invalid timestamp
        [
            VALID_SECRET,
            INVALID_HEADER_TIMESTAMP,
            VALID_PAYLOAD,
        ],
        # Test header invalid signature
        [
            VALID_SECRET,
            INVALID_HEADER_SIGNATURE,
            VALID_PAYLOAD,
        ],
        # Test invalid secret
        [
            INVALID_SECRET,
            VALID_HEADER,
            VALID_PAYLOAD,
        ],
        # Test invalid payload
        [
            VALID_SECRET,
            VALID_HEADER,
            INVALID_PAYLOAD,
        ],
    ]

    for args in invalid_inputs:
        runner = CliRunner()

        res = runner.invoke(
            verify,
            args,
        )

        assert res.exit_code == 1
        assert "ERROR" in res.output


def test_verify__valid_signature():
    """Test verify valid signature."""

    runner = CliRunner()

    res = runner.invoke(
        verify,
        [
            VALID_SECRET,
            VALID_HEADER,
            VALID_PAYLOAD,
        ],
    )

    assert res.exit_code == 0



"""Configure explorer data archive definition."""
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

import botocore
import botocore.exceptions

from ..common import (
    GencoveExplorerManager,
    request_is_from_explorer,
    validate_explorer_user_data,
)
from ....base import Command
from ....utils import user_has_aws_in_path
from .....models import ExplorerDataCredentials


class Archive(Command):
    """Archive Explorer contents command executor."""

    def __init__(self, ctx, path, credentials, options):
        super().__init__(credentials, options)
        self.ctx = ctx
        self.path = path

        # Populated after self.login() is called
        self.user = None
        self.organization = None
        self.aws_session_credentials: Optional[ExplorerDataCredentials] = None

    def validate(self):
        """Validate archive"""
        validate_explorer_user_data(self.user, self.organization)
        user_has_aws_in_path(raise_exception=True)

    def initialize(self):
        """Initialize archive subcommand."""
        self.login()
        self.user = self.api_client.get_user_details()
        self.organization = self.api_client.get_organization_details()

        if not request_is_from_explorer():
            self.aws_session_credentials = (
                self.api_client.get_explorer_data_credentials()
            )

    def execute(self):
        """Make a request to archive Explorer objects."""
        self.echo_debug("Archive Explorer contents.")

        explorer_manager = GencoveExplorerManager(
            aws_session_credentials=self.aws_session_credentials,
            user_id=str(self.user.id),
            organization_id=str(self.organization.id),
        )

        bucket, _ = (
            explorer_manager.translate_path_to_s3_path(self.path)
            .lstrip("s3://")
            .split("/", 1)
        )
        s3_client = explorer_manager.thread_safe_client("s3")

        def set_archive_tag(obj):
            """Set Archive tag to object that will be later archived
            by lifecycle rule.

            Args:
                obj (dict): Contains Key of the object to be archived.

            Returns:
                bool: False if already archived, True otherwise.
            """
            self.echo_debug(f"set_archive_tag: {obj}")

            if obj.get("StorageClass") != "STANDARD":
                return False  # skip already archived files

            try:
                s3_client.put_object_tagging(
                    Bucket=bucket,
                    Key=obj["Key"],
                    Tagging={
                        "TagSet": [
                            {
                                "Key": "Archive",
                                "Value": "DEEP_ARCHIVE",
                            }
                        ]
                    },
                )
            except botocore.exceptions.ClientError as ex:
                if (
                    ex.response
                    and ex.response.get("Error", {}).get("Code") == "AccessDenied"
                ):
                    # Check if object is already set to be archived
                    response = s3_client.get_object_tagging(
                        Bucket=bucket,
                        Key=obj["Key"],
                    )
                    archive_tag = None
                    for tag_set in response.get("TagSet", []):
                        if tag_set["Key"] == "Archive":
                            archive_tag = tag_set["Value"]
                    if archive_tag:
                        return False  # skip if file is already set to be archived
                raise

            self.echo_data(f"archive: {obj['Key']}")

            return True

        paginated_response = explorer_manager.list_s3_objects(self.path)
        obj_counts = {"archived": 0, "skipped": 0}
        with ThreadPoolExecutor() as executor:
            for response in paginated_response:
                for archived in executor.map(set_archive_tag, response["Contents"]):
                    if archived:
                        obj_counts["archived"] += 1
                    else:
                        obj_counts["skipped"] += 1

        if obj_counts["archived"]:
            self.echo_info(f"Archived {obj_counts['archived']} objects in {self.path}.")

        if obj_counts["skipped"]:
            self.echo_info(
                f"Skipped {obj_counts['skipped']} objects that were previously "
                "archived."
            )



"""Utilities for processing S3 autoimports."""


def get_line(s3_autoimport):
    """Build a list of relevant data to be printed.

    Args:
        s3_autoimport (S3ProjectImport): instance of
            S3 autoimport

    Returns:
        list(str): list of relevant data to be printed
    """
    return "\t".join(
        [
            str(s3_autoimport.id),
            str(s3_autoimport.project_id),
            str(s3_autoimport.s3_uri),
        ]
    )



"""VCR filters for upload tests."""

import copy
import json

from gencove.tests.decorators import parse_response_to_json
from gencove.tests.utils import MOCK_UUID


def filter_upload_request(request):
    """Removes destination path and aws url from request."""
    request = copy.deepcopy(request)
    if "uploads-post-data" in request.path:
        request.body = '{"destination_path": "gncv://cli-mock/test.fastq.gz"}'
    if "s3.amazonaws.com" in request.uri:
        request.uri = f"https://s3.amazonaws.com/mock_bucket/organization/{MOCK_UUID}/user/{MOCK_UUID}/uploads/{MOCK_UUID}.fastq-r1"  # noqa: E501 line too long pylint: disable=line-too-long
    return request


def filter_upload_url_request(request):
    """Removes destination path and aws url from request."""
    request = copy.deepcopy(request)
    if "uploads-url" in request.path:
        request.body = '{"destination_path": "gncv://cli-mock/test.fastq.gz", "source_url": "https://s3.amazonaws.com/example/client-id_R1.fastq.gz"}'  # noqa: E501 pylint: disable=line-too-long
    return request


def filter_project_samples_request(request):
    """Filter sample sheet sensitive data from request."""
    request = copy.deepcopy(request)
    if "project-samples" in request.path:
        try:
            body = json.loads(request.body)
            for upload in body["uploads"]:
                _filter_sample_sheet(upload)
            request.body = json.dumps(body).encode()
        except json.decoder.JSONDecodeError:
            pass
        # remove the project id from the url
        base_uri = request.uri.split("project-samples")[0]
        request.uri = base_uri + f"project-samples/{MOCK_UUID}"
    return request


@parse_response_to_json
def filter_upload_credentials_response(response, json_response):
    """Filter credentials secrets."""
    for key in ["access_key", "secret_key", "token"]:
        if key in json_response:
            json_response[key] = f"mock_{key}"
    return response, json_response


@parse_response_to_json
def filter_upload_post_data_response(response, json_response):
    """Removes sensitive data from POST to uploads-post-data."""
    if "id" in json_response:
        json_response["id"] = MOCK_UUID
    if "destination_path" in json_response:
        json_response["destination_path"] = "gncv://cli-mock/test.fastq.gz"
    if "s3" in json_response:
        json_response["s3"] = {
            "bucket": "mock_bucket",
            "object_name": f"organization/{MOCK_UUID}/user/{MOCK_UUID}/uploads/{MOCK_UUID}.fastq-r1",  # noqa: E501 pylint: disable=line-too-long
        }
    if "last_status" in json_response:
        json_response["last_status"]["id"] = MOCK_UUID
    return response, json_response


@parse_response_to_json
def filter_import_fastqs_from_url_response(response, json_response):
    """Removes sensitive data from POST to uploads-post-data."""
    if "id" in json_response:
        json_response["id"] = MOCK_UUID
    if "destination_path" in json_response:
        json_response["destination_path"] = "gncv://cli-mock/test.fastq.gz"
    if "s3" in json_response:
        json_response["s3"] = {
            "bucket": "mock_bucket",
            "object_name": f"organization/{MOCK_UUID}/user/{MOCK_UUID}/uploads/{MOCK_UUID}.fastq-r1",  # noqa: E501 pylint: disable=line-too-long
        }
    if "last_status" in json_response:
        json_response["last_status"]["id"] = MOCK_UUID
    if "source_url" in json_response:
        json_response[
            "source_url"
        ] = "https://s3.amazonaws.com/example/client-id_R1.fastq.gz"
    return response, json_response


@parse_response_to_json
def filter_volatile_dates(response, json_response):
    """Patches volatile dates, and push them into the future, otherwise
    the app will think the token/whatever is expired.
    """
    if "expiry_time" in json_response:
        json_response["expiry_time"] = "2050-09-03T00:00:00+00:00"
    return response, json_response


def _filter_sample_sheet(result):
    """Common function that filters sample sheet sensitive data."""
    if "client_id" in result:
        result["client_id"] = "mock-client-id"
    if "fastq" in result and "r1" in result["fastq"]:
        r1 = result["fastq"]["r1"]  # pylint: disable=invalid-name
        if "upload" in r1:
            r1["upload"] = MOCK_UUID
        if "destination_path" in r1:
            r1["destination_path"] = "gncv://cli-mock/test.fastq.gz"
        if "last_status" in r1:
            r1["last_status"]["id"] = MOCK_UUID
        if "last_status" in r1:
            r1["last_status"]["id"] = MOCK_UUID
    if "sample" in result:
        result["sample"] = MOCK_UUID


@parse_response_to_json
def filter_sample_sheet_response(response, json_response):
    """Filter sample sheet sensitive data from response."""
    if "results" in json_response:
        for result in json_response["results"]:
            _filter_sample_sheet(result)
    return response, json_response


@parse_response_to_json
def filter_project_samples_response(response, json_response):
    """Filter sample sheet sensitive data from upload response."""
    if "uploads" in json_response:
        for upload in json_response["uploads"]:
            _filter_sample_sheet(upload)
    return response, json_response



"""VCR filters for explorer tests."""

import json
import re

from gencove.tests.decorators import parse_response_to_json
from gencove.tests.utils import MOCK_UUID


@parse_response_to_json
def filter_list_instances_response(response, json_response):
    """Filter list instances sensitive data from response."""
    if "results" in json_response:
        for result in json_response["results"]:
            if "id" in result:
                result["id"] = MOCK_UUID
    if "instance_ids" in json_response:
        messages = []
        for msg in json_response["instance_ids"]:
            # Replace UUIDs with a MOCK_UUID string
            uuid_pattern = re.compile(
                r"\b[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89a-f][a-f0-9]{3}-[a-f0-9]{12}\b",  # noqa: E501, pylint: disable=line-too-long
                re.IGNORECASE,
            )
            messages.append(uuid_pattern.sub(MOCK_UUID, msg))
        json_response["instance_ids"] = messages

    return response, json_response


def filter_instance_ids_request(request):
    """Filter sensitive data from requests that have instance_ids."""
    if request.body is None:
        return request
    try:
        body = json.loads(request.body)
        if "instance_ids" in body:
            samples = [MOCK_UUID for _ in body["instance_ids"]]
            body["instance_ids"] = samples
            request.body = json.dumps(body).encode()
        if "instance_id" in body:
            body["instance_id"] = MOCK_UUID
            request.body = json.dumps(body).encode()
    except json.decoder.JSONDecodeError:
        pass
    return request


@parse_response_to_json
def filter_session_credentials_response(response, json_response):
    """Filter credentials secrets."""
    for key in [
        "access_key",
        "secret_key",
        "token",
        "ec2_instance_id",
        "ssm_document_name",
    ]:
        if key in json_response:
            json_response[key] = f"mock_{key}"
    return response, json_response


@parse_response_to_json
def filter_data_credentials_response(response, json_response):
    """Filter credentials secrets."""
    for key in ["access_key", "secret_key", "token"]:
        if key in json_response:
            json_response[key] = f"mock_{key}"
    if "id" in json_response:
        json_response["id"] = MOCK_UUID
    if "results" in json_response:
        for result in json_response["results"]:
            if "id" in result:
                result["id"] = MOCK_UUID
    if "roles" in json_response:
        json_response["roles"]["organization"]["id"] = MOCK_UUID
    if "email" in json_response:
        json_response["email"] = "email@example.com"
    return response, json_response



"""Test project's get reference genome command."""
# pylint: disable=import-error, wrong-import-order

import os
from unittest.mock import call
from uuid import uuid4

from click.testing import CliRunner

from gencove.command.download.utils import download_file
from gencove.command.projects.get_reference_genome.cli import get_reference_genome
from gencove.models import SampleFile
from gencove.tests.decorators import assert_authorization, assert_no_requests
from gencove.tests.filters import (
    filter_aws_headers,
    filter_file_types_request,
    filter_jwt,
    mock_binary_response,
    replace_gencove_url_vcr,
    replace_s3_from_url,
)
from gencove.tests.projects.vcr.filters import (
    filter_projects_detail_request,
    filter_projects_detail_response,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import MOCK_UUID

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr/get_reference_genome",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_file_types_request,
            filter_projects_detail_request,
            replace_s3_from_url,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
            mock_binary_response,
            filter_aws_headers,
            filter_projects_detail_response,
        ],
    }


@assert_no_requests
def test_no_required_options(credentials, mocker):  # pylint: disable=unused-argument
    """Test that command exits without project id provided."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        res = runner.invoke(
            get_reference_genome,
            [
                "22",
                "cli_test_data",
                *credentials,
            ],
        )
        assert res.exit_code == 1
        assert "ERROR: Project ID is not valid. Exiting." in res.output


@pytest.mark.vcr
@assert_authorization
def test_get_reference_genome__not_owned_project(
    credentials, mocker
):  # pylint: disable=unused-argument
    """Test get reference genome fails when project is not owned."""
    runner = CliRunner()
    #
    project_id = str(uuid4())
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        res = runner.invoke(
            get_reference_genome,
            [
                project_id,
                "cli_test_data",
                *credentials,
            ],
        )
    assert res.exit_code == 1
    assert "WARNING: There was an error while validating file types." in res.output
    assert f"ERROR: Project {project_id} does not exist." in res.output


@pytest.mark.vcr
@assert_authorization
@pytest.mark.parametrize("remove_hyphens", [True, False])
def test_get_reference_genome__success(  # pylint: disable=too-many-arguments
    credentials, mocker, project_id_reference_genome, recording, remove_hyphens
):
    """Test get reference genome success."""
    mocked_download_file = mocker.patch(
        "gencove.command.projects.get_reference_genome.main.download_file",
        side_effect=download_file,
    )
    runner = CliRunner()
    if remove_hyphens:
        project_id_reference_genome = project_id_reference_genome.replace("-", "")
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        res = runner.invoke(
            get_reference_genome,
            [
                project_id_reference_genome,
                "cli_test_data",
                *credentials,
            ],
        )
        assert res.exit_code == 0
    assert mocked_download_file.call_count == 10
    if not recording:
        file_types = [
            ("genome-dict", "dict"),
            ("genome-fasta", "fasta.gz"),
            ("genome-fasta_amb", "fasta.gz.amb"),
            ("genome-fasta_ann", "fasta.gz.ann"),
            ("genome-fasta_bwt", "fasta.gz.bwt"),
            ("genome-fasta_fai", "fasta.gz.fai"),
            ("genome-fasta_gzi", "fasta.gz.gzi"),
            ("genome-fasta_pac", "fasta.gz.pac"),
            ("genome-fasta_sa", "fasta.gz.sa"),
            ("genome-fasta_vcf_header", "fasta.gz.vcf_header"),
        ]

        calls = []
        for file_type, extension in file_types:
            filename = f"{MOCK_UUID}_{file_type}.{extension}"
            query_param = (
                f"response-content-disposition=attachment%3B+filename%3D{filename}"
            )
            download_url = f"https://example.com/genome.{extension}?{query_param}"
            calls.append(
                call(
                    f"cli_test_data/{filename}",
                    # We need to use SampleFile to have a full HttpUrl pydantic
                    # attribute
                    SampleFile(id=MOCK_UUID, download_url=download_url).download_url,
                    no_progress=False,
                )
            )
        mocked_download_file.assert_has_calls(calls, any_order=True)


# pylint: disable=too-many-arguments
@pytest.mark.vcr
@assert_authorization
def test_get_reference_genome__success_specific_file_type(
    credentials, mocker, project_id_reference_genome, recording
):
    """Test get reference genome success specifying file types."""
    mocked_download_file = mocker.patch(
        "gencove.command.projects.get_reference_genome.main.download_file",
        side_effect=download_file,
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        res = runner.invoke(
            get_reference_genome,
            [
                project_id_reference_genome,
                "cli_test_data",
                "--file-types",
                "genome-dict",
                *credentials,
            ],
        )
        assert res.exit_code == 0
    assert mocked_download_file.call_count == 1
    if not recording:
        filename = f"{MOCK_UUID}_genome-dict.dict"
        query_param = (
            f"response-content-disposition=attachment%3B+filename%3D{filename}"
        )
        download_url = f"https://example.com/genome.dict?{query_param}"
        mocked_download_file.assert_called_once_with(
            f"cli_test_data/{filename}",
            SampleFile(
                id=MOCK_UUID,
                download_url=download_url,
            ).download_url,
            no_progress=False,
        )


# pylint: disable=too-many-arguments
@pytest.mark.vcr
@assert_authorization
def test_get_reference_genome__success_no_progress(
    credentials, mocker, project_id_reference_genome, recording
):
    """Test command doesn't show progress bar."""
    mocked_download_file = mocker.patch(
        "gencove.command.projects.get_reference_genome.main.download_file",
        side_effect=download_file,
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        res = runner.invoke(
            get_reference_genome,
            [
                project_id_reference_genome,
                "cli_test_data",
                "--file-types",
                "genome-dict",
                *credentials,
                "--no-progress",
            ],
        )
        assert res.exit_code == 0
    assert mocked_download_file.call_count == 1
    if not recording:
        filename = f"{MOCK_UUID}_genome-dict.dict"
        query_param = (
            f"response-content-disposition=attachment%3B+filename%3D{filename}"
        )
        download_url = f"https://example.com/genome.dict?{query_param}"
        mocked_download_file.assert_called_once_with(
            f"cli_test_data/{filename}",
            SampleFile(
                id=MOCK_UUID,
                download_url=download_url,
            ).download_url,
            no_progress=True,
        )


@pytest.mark.vcr
@assert_authorization
def test_get_reference_genome__bad_file_type(  # pylint: disable=too-many-arguments
    credentials, mocker, project_id_reference_genome
):
    """Fails when an invalid file type is provided."""
    mocked_download_file = mocker.patch(
        "gencove.command.projects.get_reference_genome.main.download_file",
        side_effect=download_file,
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        res = runner.invoke(
            get_reference_genome,
            [
                project_id_reference_genome,
                "cli_test_data",
                "--file-types",
                "bad-file-type",
                *credentials,
            ],
        )
        assert res.exit_code == 1
        assert "Invalid file types: bad-file-type." in res.output
        assert (
            "Run 'gencove file-types --object reference-genome' command for a list of "
            "valid file types. " in res.output
        )
    mocked_download_file.assert_not_called()



"""Request to delete projects"""
from gencove import client  # noqa: I100
from gencove.command.base import Command


class Delete(Command):
    """Delete projects."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        project_ids,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.project_ids = project_ids

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """

    def execute(self):
        """Make a request to request delete projects."""
        self.echo_debug(f"Requesting to delete projects {self.project_ids}")

        try:
            deleted_projects_details = self.api_client.delete_projects(
                project_ids=self.project_ids,
            )

            self.echo_debug(deleted_projects_details)
            self.echo_info("The following projects have been deleted successfully:")
            for project in self.project_ids:
                self.echo_info(f"\t{project}")
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning("There was an error requesting delete projects.")
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            elif err.status_code == 404:
                self.echo_warning(
                    f"Some of the projects {self.project_ids} do not exist."
                )
            else:
                raise



"""Test reports project-qc command."""

# pylint: disable=wrong-import-order, import-error
import operator
import os
from pathlib import Path

from click.testing import CliRunner

from gencove.client import APIClient, APIClientError  # noqa: I100
from gencove.command.reports.cli import project_qc
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import (
    filter_aws_headers,
    filter_jwt,
    replace_gencove_url_vcr,
)
from gencove.tests.reports.vcr.filters import (
    filter_project_qc_report_request,
    filter_report_response_body,
    filter_report_response_filename,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_response_from_vcr_dict, get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/reports/vcr",
        "filter_headers": ["Authorization", "Content-Length", "User-Agent", "ETag"],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_project_qc_report_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_aws_headers,
            filter_volatile_dates,
            filter_report_response_body,
            filter_report_response_filename,
        ],
    }


@pytest.mark.vcr
@assert_authorization
def test_project_qc__not_owned(
    credentials, mocker, recording, vcr
):  # pylint: disable=unused-argument
    """Test project qc report not owned case"""
    runner = CliRunner()
    if not recording:
        project_qc_dict = get_vcr_response(
            "/api/v2/project-qc-report/", vcr, operator.contains, just_body=False
        )
        response = get_response_from_vcr_dict(project_qc_dict)

        mocked_project_qc = mocker.patch.object(
            APIClient,
            "get_project_qc_report",
            side_effect=APIClientError(message="", status_code=response.status_code),
            return_value=response,
        )
    with runner.isolated_filesystem():
        os.mkdir("tempdir")
        res = runner.invoke(
            project_qc,
            [
                "11111111-1111-1111-1111-111111111111",
                *credentials,
            ],
        )

    assert res.exit_code == 0
    if not recording:
        mocked_project_qc.assert_called_once()
    assert "does not exist or you do not have access" in res.output


@pytest.mark.vcr
@assert_authorization
def test_project_qc__success(  # pylint: disable=too-many-arguments,unused-argument
    credentials, mocker, project_id, recording, vcr
):
    """Test QC report success case"""
    runner = CliRunner()
    if not recording:
        project_qc_dict = get_vcr_response(
            "/api/v2/project-qc-report/", vcr, operator.contains, just_body=False
        )
        response = get_response_from_vcr_dict(project_qc_dict)

        # Need to reconstruct the raw response
        mocked_project_qc = mocker.patch.object(
            APIClient,
            "get_project_qc_report",
            return_value=response,
        )
    with runner.isolated_filesystem():
        res = runner.invoke(
            project_qc,
            [
                project_id,
                *credentials,
            ],
        )
        csv_file = Path("report.csv")  # filename from mocked response
        assert csv_file.exists()
        csv_contents = csv_file.read_text(encoding="utf-8")
        assert csv_contents == response.text

    assert res.exit_code == 0
    if not recording:
        mocked_project_qc.assert_called_once()
    assert "Saved project QC report CSV" in res.output


@pytest.mark.vcr
@assert_authorization
def test_project_qc__success_columns(
    credentials, mocker, project_id, recording, vcr
):  # pylint: disable=too-many-arguments,too-many-locals,unused-argument
    """Test QC report success case with requested coluns"""
    runner = CliRunner()
    if not recording:
        project_qc_dict = get_vcr_response(
            "/api/v2/project-qc-report/", vcr, operator.contains, just_body=False
        )
        response = get_response_from_vcr_dict(project_qc_dict)

        # Need to reconstruct the raw response
        mocked_project_qc = mocker.patch.object(
            APIClient,
            "get_project_qc_report",
            return_value=response,
        )
    with runner.isolated_filesystem():
        os.mkdir("tempdir")
        outfile = "tempdir/test.csv"
        res = runner.invoke(
            project_qc,
            [
                "--output-filename",
                "tempdir/test.csv",
                "--columns",
                "id,client_id",
                project_id,
                *credentials,
            ],
        )
        with open(outfile, "r", encoding="utf-8") as fileobj:
            contents = fileobj.readlines()

    assert res.exit_code == 0
    if not recording:
        mocked_project_qc.assert_called_once()
    assert "Saved project QC report CSV" in res.output

    # Confirm columns are as expected
    columns_row = contents[0]
    columns = columns_row.strip().split(",")
    assert len(columns) == 2
    assert columns[0] == "id"
    assert columns[1] == "client_id"


@pytest.mark.vcr
@assert_authorization
def test_project_qc__bad_columns(  # pylint: disable=too-many-arguments,unused-argument
    credentials, mocker, project_id, recording, vcr
):
    """Test QC report with bad column name passed to --columns"""
    runner = CliRunner()
    if not recording:
        project_qc_dict = get_vcr_response(
            "/api/v2/project-qc-report/", vcr, operator.contains, just_body=False
        )
        response = get_response_from_vcr_dict(project_qc_dict)

        # Need to reconstruct the raw response
        mocked_project_qc = mocker.patch.object(
            APIClient,
            "get_project_qc_report",
            side_effect=APIClientError(
                message=response.content,
                status_code=response.status_code,
            ),
            return_value=response,
        )

    with runner.isolated_filesystem():
        res = runner.invoke(
            project_qc,
            [
                "--columns",
                "id,invalid_column",
                project_id,
                *credentials,
            ],
        )

    assert res.exit_code == 0
    if not recording:
        mocked_project_qc.assert_called_once()
    assert "There was an error retrieving the project QC report" in res.output
    assert "Acceptable column names include the following" in res.output



"""Constants for basespace-import command."""
from typing import Optional

from gencove.constants import Optionals


class BaseSpaceImportOptionals(Optionals):
    """BaseSpaceImportOptionals model"""

    metadata_json: Optional[str]



"""Test cases for base command class."""
# pylint: disable=wrong-import-order,import-error,unused-import

from gencove.command.base import Command
from gencove.constants import Credentials, Optionals

import pytest  # noqa: F401


class SimpleCommand(Command):
    """Dummy command for testing, does not extend base class functionality"""

    def initialize(self):
        """Initialize command"""
        pass  # pylint: disable=unnecessary-pass

    def validate(self):
        """Validate command"""
        pass  # pylint: disable=unnecessary-pass

    def execute(self):
        """Execute command"""
        pass  # pylint: disable=unnecessary-pass


class TestFetchCredentialsFromEnv:
    """Test class for fetching credentials from environment variable behavior"""

    def test_no_credentials_provided_uses_api_env_var(
        self, monkeypatch
    ):  # pylint: disable=redefined-outer-name,no-self-use
        """Test Command reads API key from env"""
        monkeypatch.setenv("GENCOVE_API_KEY", "env_api_key")

        empty_credentials = Credentials(api_key="", email="", password="")

        command = SimpleCommand(
            credentials=empty_credentials, options=Optionals(host="http://example.com")
        )

        # Command should prioritize API key from environment variable, ignore email
        assert command.credentials.api_key == "env_api_key"
        assert command.credentials.password == ""
        assert command.credentials.email == ""
        assert command.is_credentials_valid is True

    def test_api_key_provided_skips_env_vars(
        self, monkeypatch
    ):  # pylint: disable=redefined-outer-name,no-self-use
        """Test Command uses provided API key and ignores ENV variables"""
        monkeypatch.setenv("GENCOVE_API_KEY", "wrong_api_key")

        credentials = Credentials(api_key="correct_api_key", email="", password="")
        command = SimpleCommand(
            credentials=credentials, options=Optionals(host="http://example.com")
        )

        assert command.credentials.api_key == "correct_api_key"
        assert command.credentials.email == ""
        assert command.credentials.password == ""
        assert command.is_credentials_valid is True

    def test_email_password_provided_skips_env_vars(
        self, monkeypatch
    ):  # pylint: disable=redefined-outer-name,no-self-use
        """Test Command uses provided email and password and ignores ENV variables"""
        monkeypatch.setenv("GENCOVE_EMAIL", "wrong_email@example.com")
        monkeypatch.setenv("GENCOVE_PASSWORD", "wrong_password")

        credentials = Credentials(
            api_key="", email="correct_email@example.com", password="correct_password"
        )
        command = SimpleCommand(
            credentials=credentials, options=Optionals(host="http://example.com")
        )

        assert command.credentials.api_key == ""
        assert command.credentials.email == "correct_email@example.com"
        assert command.credentials.password == "correct_password"
        assert command.is_credentials_valid is True

    def test_no_env_vars_empty_defaults(
        self,
    ):  # pylint: disable=redefined-outer-name,no-self-use
        """Test Command uses empty strings for credentials when no ENV variables set and
        no creds provided

        In this case, user is prompted for email and password."""
        empty_credentials = Credentials(api_key="", email="", password="")
        command = SimpleCommand(
            credentials=empty_credentials, options=Optionals(host="http://example.com")
        )

        assert command.credentials.api_key == ""
        assert command.credentials.email == ""
        assert command.credentials.password == ""
        assert command.is_credentials_valid is True

    def test_email_uses_env_password(
        self,
        monkeypatch,
    ):  # pylint: disable=redefined-outer-name,no-self-use
        """Test Command uses provided email and ENV password"""
        monkeypatch.setenv("GENCOVE_EMAIL", "env_email@example.com")
        monkeypatch.setenv("GENCOVE_PASSWORD", "env_password")

        credentials = Credentials(
            api_key="", email="provided_email@example.com", password=""
        )
        command = SimpleCommand(
            credentials=credentials, options=Optionals(host="http://example.com")
        )

        assert command.credentials.api_key == ""
        assert command.credentials.email == "provided_email@example.com"
        assert command.credentials.password == "env_password"
        assert command.is_credentials_valid is True

    def test_multiple_credentials_not_allowed(
        self, capsys
    ):  # pylint: disable=redefined-outer-name,no-self-use
        """Test Command does not allow multiple credentials to be provided directly"""
        credentials = Credentials(
            api_key="api_key", email="email@example.com", password="password"
        )
        command = SimpleCommand(
            credentials=credentials, options=Optionals(host="http://example.com")
        )
        captured = capsys.readouterr()

        assert command.is_credentials_valid is False
        assert "Multiple sets of credentials provided." in captured.err

    def test_multiple_credentials_via_env_not_allowed(
        self, monkeypatch, capsys
    ):  # pylint: disable=redefined-outer-name,no-self-use
        """Test Command does not allow multiple credentials to be provided through env
        vars"""
        credentials = Credentials(api_key="", email="", password="")
        monkeypatch.setenv("GENCOVE_API_KEY", "env_api_key")
        monkeypatch.setenv("GENCOVE_EMAIL", "env_email@example.com")

        command = SimpleCommand(
            credentials=credentials, options=Optionals(host="http://example.com")
        )
        captured = capsys.readouterr()

        assert command.is_credentials_valid is False
        assert (
            "Multiple sets of credentials provided via environment variables."
            in captured.err
        )



"""Constants for s3-autoimport command."""
from typing import Optional

from gencove.constants import Optionals


class S3AutoImportOptionals(Optionals):
    """S3AutoImportOptionals model"""

    metadata_json: Optional[str]



"""Test available pipeline capabilities list command."""
# pylint: disable=wrong-import-order, import-error
import io
import operator
import sys
import uuid

from click.testing import CliRunner


from gencove.client import APIClient, APIClientError, APIClientTimeout
from gencove.command.projects.cli import list_project_pipeline_capabilities
from gencove.logger import echo_data
from gencove.models import PipelineDetail
from gencove.tests.decorators import assert_authorization, assert_no_requests
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.projects.vcr.filters import (
    filter_project_pipeline_capabilities_request,
    filter_project_pipeline_capabilities_response,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_project_pipeline_capabilities_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
            filter_project_pipeline_capabilities_response,
        ],
    }


@assert_no_requests
def test_list_pipeline_capabilities__bad_pipeline_id(
    mocker, credentials
):  # pylint: disable=unused-argument
    """Test list pipeline capabilities throw an error if pipeline id is not uuid."""
    runner = CliRunner()

    pipeline_id = "123456"

    res = runner.invoke(
        list_project_pipeline_capabilities,
        [pipeline_id, *credentials],
    )
    assert res.exit_code == 1
    output_line = io.BytesIO()
    sys.stdout = output_line
    echo_data(
        "\n".join(
            [
                "ERROR: Pipeline ID is not valid. Exiting.",
                "Aborted!",
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_pipeline_capabilities__no_pipeline(mocker, credentials):
    """Test list pipeline capabilities throw an error if no pipeline available."""
    runner = CliRunner()
    mocked_get_pipeline_capabilities_for_pipeline = mocker.patch.object(
        APIClient,
        "get_pipeline_capabilities_for_pipeline",
        side_effect=APIClientError(
            message="API Client Error: Not Found: Not found.", status_code=404
        ),
        return_value={"detail": "Not found"},
    )

    pipeline_id = str(uuid.uuid4())

    res = runner.invoke(
        list_project_pipeline_capabilities,
        [pipeline_id, *credentials],
    )
    assert res.exit_code == 1
    mocked_get_pipeline_capabilities_for_pipeline.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo_data(
        "\n".join(
            [
                f"ERROR: Pipeline {pipeline_id} does not exist.",
                "ERROR: API Client Error: Not Found: Not found.",
                "Aborted!",
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()


@pytest.mark.vcr
@assert_authorization
def test_list_pipeline_capabilities__empty(
    credentials, pipeline_id, mocker, recording, vcr
):
    """Test there are no pipeline capabilities."""
    # WHEN UPDATING THE CASSETTE, CHANGE THE FILE SO IT HAS AN EMPTY CAPABILITIES LIST
    runner = CliRunner()
    if not recording:
        get_pipeline_response = get_vcr_response(
            "/api/v2/pipeline/", vcr, operator.contains
        )
        mocked_get_pipeline_capabilities_for_pipeline = mocker.patch.object(
            APIClient,
            "get_pipeline_capabilities_for_pipeline",
            return_value=PipelineDetail(**get_pipeline_response),
        )
    res = runner.invoke(
        list_project_pipeline_capabilities,
        [pipeline_id, *credentials],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_get_pipeline_capabilities_for_pipeline.assert_called_once()
    assert res.output == ""


@pytest.mark.vcr
@assert_authorization
def test_list_pipeline_capabilities__not_empty(
    mocker, credentials, pipeline_id, recording, vcr
):
    """Test pipeline capabilities being output to the shell."""
    runner = CliRunner()
    if not recording:
        # Mock get_pipeline_capabilities_for_pipeline only if using the cassettes,
        # since we mock the return value.
        get_pipeline_response = get_vcr_response(
            "/api/v2/pipeline/", vcr, operator.contains
        )
        mocked_get_pipeline_capabilities_for_pipeline = mocker.patch.object(
            APIClient,
            "get_pipeline_capabilities_for_pipeline",
            return_value=PipelineDetail(**get_pipeline_response),
        )
    res = runner.invoke(
        list_project_pipeline_capabilities,
        [pipeline_id, *credentials],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_get_pipeline_capabilities_for_pipeline.assert_called_once()
        pipelines = "\n".join(
            [
                f"{capability['id']}\t{capability['name']}"
                for capability in get_pipeline_response["capabilities"]
            ]
        )
        assert res.output.startswith(pipelines) is True


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_pipeline_capabilities__not_empty_slow_response_retry(
    credentials, pipeline_id, mocker
):
    """Test pipeline capabilities slow response retry."""
    runner = CliRunner()
    mocked_get_pipeline_capabilities_for_pipeline = mocker.patch.object(
        APIClient,
        "get_pipeline_capabilities_for_pipeline",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )

    res = runner.invoke(
        list_project_pipeline_capabilities,
        [pipeline_id, *credentials],
    )
    assert res.exit_code == 1
    assert mocked_get_pipeline_capabilities_for_pipeline.call_count == 2



"""Common command line options.

These options must be added to all commands.
"""
import os

import click

from gencove.constants import HOST

common_options = [  # pylint: disable=invalid-name
    click.option(
        "--host",
        default=lambda: os.environ.get("GENCOVE_HOST", HOST),
        help="Optional Gencove API host, including http/s protocol. "
        "Can be passed as GENCOVE_HOST environment variable. "
        f"Defaults to {HOST}",
    ),
    click.option(
        "--email",
        default="",
        help="Gencove user email to be used in login. "
        "Can be passed as GENCOVE_EMAIL environment variable.",
    ),
    click.option(
        "--password",
        default="",
        help="Gencove user password to be used in login. "
        "Can be passed as GENCOVE_PASSWORD environment variable. "
        "When MFA is configured, an MFA token will have to be provided after "
        "the command is executed. Only used if --email is provided.",
    ),
    click.option(
        "--api-key",
        default="",
        help="Gencove api key. "
        "Can be passed as GENCOVE_API_KEY environment variable. "
        "When using the API key, an MFA token does not need to be provided.",
    ),
]


def add_options(options):  # pylint: disable=missing-function-docstring
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options



"""Monthly usage report executor."""
from pathlib import Path

from gencove import client  # noqa: I100
from gencove.command.base import Command
from gencove.command.utils import extract_filename_from_headers


class MonthlyUsageReport(Command):
    """Monthly usage report executor."""

    def __init__(  # pylint: disable=too-many-arguments,invalid-name
        self,
        from_,
        to,
        output_filename,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.from_ = from_
        self.to = to
        self.output_filename = output_filename

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """

    def execute(self):
        """Request to get project QC report and download it."""
        self.echo_info("Retrieving monthly usage report for organization")

        try:
            monthly_usage_report = (
                self.api_client.get_organization_monthly_usage_report(
                    from_=self.from_, to=self.to
                )
            )

            if not self.output_filename:
                filename = extract_filename_from_headers(
                    headers=monthly_usage_report.headers
                )
                if not filename:
                    filename = "monthly-usage-report.csv"
                self.output_filename = Path.cwd() / filename
            else:
                Path(self.output_filename).parent.mkdir(exist_ok=True, parents=True)

            with open(self.output_filename, "wb") as filename:
                filename.write(monthly_usage_report.content)
            self.echo_info(
                f"Saved organization monthly usage report CSV "
                f"to {Path(self.output_filename).resolve()}"
            )
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning(
                    "There was an error retrieving the monthly usage report."
                )
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            elif err.status_code == 404:
                self.echo_warning("You do not have access to this report")
            else:
                raise



"""Test decorators."""

import json
import os
from functools import wraps

import requests


def parse_response_to_json(func):
    """Decorator that parses the response body into JSON and passes it to the
    function, then it saves the changes into the response body again.
    """

    @wraps(func)
    def wrapper(response):
        try:
            json_response = json.loads(response["body"]["string"])
        except (json.decoder.JSONDecodeError, UnicodeDecodeError):
            json_response = {}
        response, json_response = func(response, json_response)
        if response is not None and json_response:
            response["body"]["string"] = json.dumps(json_response).encode()
        return response

    return wrapper


def assert_authorization(func):
    """Decorator that ensures that the API key is being passed to our
    endpoints, and the login endpoint is called if email/password is provided.
    """

    @wraps(func)
    def wrapper(*args, mocker, **kwargs):
        host = os.getenv("GENCOVE_HOST")
        api_key = os.getenv("GENCOVE_API_KEY_TEST")
        email = os.getenv("GENCOVE_EMAIL_TEST")
        password = os.getenv("GENCOVE_PASSWORD_TEST")
        using_api_key = os.getenv("USING_API_KEY")

        def mock_get_auth(url, *args, headers, **kwargs):
            if host in url and using_api_key:
                assert (
                    headers["Authorization"] == f"Api-Key {api_key}"
                ), f"No valid authorization header provided for GET-{url}"
            kwargs["headers"] = headers
            return requests.get(url, *args, **kwargs)

        mocker.patch("gencove.client.get", side_effect=mock_get_auth)
        login_called = False

        def mock_post_auth(url, data, *args, headers, **kwargs):
            nonlocal login_called
            if host in url and using_api_key:
                assert (
                    headers["Authorization"] == f"Api-Key {api_key}"
                ), f"No valid authorization header provided for POST-{url}"
            elif host in url:
                if url.endswith("jwt-create/"):
                    login_called = True
                    data_json = json.loads(data)
                    assert data_json["email"] == email
                    assert data_json["password"] == password
                else:
                    assert "Bearer " in headers.get("Authorization")
            kwargs["headers"] = headers
            return requests.post(url, data, *args, **kwargs)

        mocker.patch("gencove.client.post", side_effect=mock_post_auth)
        kwargs["mocker"] = mocker
        func(*args, **kwargs)
        if not using_api_key:
            assert login_called, "jwt-create endpoint was not called"

    return wrapper


def assert_no_requests(func):
    """Decorator that mocks get, post and delete and asserts that
    none of those methods are called
    """

    @wraps(func)
    def wrapper(*args, mocker, **kwargs):
        mock_get = mocker.patch("gencove.client.get")
        mock_post = mocker.patch("gencove.client.post")
        mock_delete = mocker.patch("gencove.client.delete")

        kwargs["mocker"] = mocker
        func(*args, **kwargs)

        mock_get.assert_not_called()
        mock_post.assert_not_called()
        mock_delete.assert_not_called()

    return wrapper



"""List BaseSpace projects shell command definition.
"""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import BaseSpaceList


@click.command("list")
@add_options(common_options)
def basespace_list(
    host,
    email,
    password,
    api_key,
):
    """List all BaseSpace projects.

    Examples:

        Import Biosamples to a project:

            gencove basespace projects list

    """
    BaseSpaceList(
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Commands to be executed from command line."""
# pylint: disable=E0012,C0330,R0913
import click

from .monthly_usage.cli import monthly_usage
from .project_qc.cli import project_qc


@click.group()
def reports():
    """Report management commands."""


reports.add_command(project_qc)
reports.add_command(monthly_usage)



"""Test instances stop command."""
# pylint: disable=wrong-import-order, import-error
import io
import sys


from click import echo
from click.testing import CliRunner

from gencove.client import (
    APIClient,
    APIClientError,
)  # noqa: I100
from gencove.command.explorer.instances.cli import stop
from gencove.models import (
    ExplorerInstance,
    ExplorerInstances,
    ResponseMeta,
)
from gencove.tests.decorators import assert_authorization
from gencove.tests.explorer.vcr.filters import (  # noqa: I101
    filter_list_instances_response,
    filter_instance_ids_request,
)
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import MOCK_UUID, get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/explorer/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_instance_ids_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
            filter_list_instances_response,
        ],
    }


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_instances_stop_no_permission(mocker, credentials):
    """Test instances no permission available to stop them."""
    runner = CliRunner()
    mocked_get_instances = mocker.patch.object(
        APIClient,
        "get_explorer_instances",
        side_effect=APIClientError(
            message="API Client Error: Not Found: Not found.", status_code=403
        ),
        return_value={"detail": "Not found"},
    )
    res = runner.invoke(stop, credentials)
    assert res.exit_code == 1
    mocked_get_instances.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\n".join(
            [
                "ERROR: You do not have the sufficient permission "
                "level required to perform this operation.",
                "ERROR: API Client Error: Not Found: Not found.",
                "Aborted!",
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()


@pytest.mark.vcr
@assert_authorization
def test_instances_stop(mocker, credentials, recording, vcr):
    """Test instances being stopped."""
    runner = CliRunner()
    if not recording:
        # Mock stop only if using the cassettes, since we mock the
        # return value.
        get_vcr_response("/api/v2/explorer-stop-instances/", vcr)
        mocked_instances_stop = mocker.patch.object(
            APIClient,
            "stop_explorer_instances",
            return_value=None,
        )
    res = runner.invoke(stop, credentials)
    assert b"Request to stop explorer instances accepted." in res.output.encode()
    assert res.exit_code == 0
    if not recording:
        mocked_instances_stop.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_instances_stop_already_stopped(mocker, credentials, recording, vcr):
    """Test instances already being stopped."""
    runner = CliRunner()
    if not recording:
        # Mock stop only if using the cassettes, since we mock the
        # return value.
        response_json = get_vcr_response("/api/v2/explorer-stop-instances/", vcr)
        error_msg = "\n".join(
            [
                f"  {key}: {value[0] if isinstance(value, list) else str(value)}"  # noqa: E501  # pylint: disable=line-too-long
                for key, value in response_json.items()
            ]
        )
        mocked_instances_stop = mocker.patch.object(
            APIClient,
            "stop_explorer_instances",
            side_effect=APIClientError(
                message=f"API Client Error: Bad Request:\n{error_msg}",
                status_code=400,
            ),
        )
    res = runner.invoke(stop, credentials)
    assert b"already in stopped status" in res.output.encode()
    assert res.exit_code == 1
    if not recording:
        mocked_instances_stop.assert_called_once()


@pytest.mark.vcr
@pytest.mark.default_cassette("jwt-create.yaml")
@assert_authorization
def test_instances_stop_update_cli(mocker, credentials):
    """Test instances being not stopped because CLI is outdated."""
    runner = CliRunner()
    mocked_get_instances = mocker.patch.object(
        APIClient,
        "get_explorer_instances",
        return_value=ExplorerInstances(
            meta=ResponseMeta(count=2, next=None, previous=None),
            results=[
                ExplorerInstance(
                    id=MOCK_UUID, status="running", stop_after_inactivity_hours=None
                ),
                ExplorerInstance(
                    id=MOCK_UUID, status="running", stop_after_inactivity_hours=None
                ),
            ],
        ),
    )
    res = runner.invoke(stop, credentials)
    assert (
        b"Command not supported. Download the latest version of the Gencove CLI."
        in res.output.encode()
    )
    assert res.exit_code == 1
    mocked_get_instances.assert_called_once()



"""Start shell session for explorer instance subcommand."""

import base64
import os
import signal
import subprocess  # nosec B404 (bandit subprocess import)
import sys
import time
from multiprocessing import Process

from gencove.utils import get_boto_session_refreshable

from ....base import Command
from ....utils import (
    user_has_aws_in_path,
    user_has_session_manager_plugin_in_path,
    user_has_supported_aws_cli,
)
from .....exceptions import ValidationError
from .....models import ExplorerShellSessionCredentials


class ShellSession(Command):
    """Start shell session for explorer instance command executor."""

    def validate(self):
        """Validate start shell sessions"""
        user_has_aws_in_path(raise_exception=True)
        user_has_supported_aws_cli(raise_exception=True)
        user_has_session_manager_plugin_in_path(raise_exception=True)

    def initialize(self):
        """Initialize shell subcommand."""
        self.login()

    def execute(self):
        """Make a request to get shell session credentials."""

        explorer_instances = self.api_client.get_explorer_instances()
        self.echo_debug(f"Found {len(explorer_instances.results)} explorer instances.")

        if len(explorer_instances.results) != 1:
            raise ValidationError(
                "Command not supported. Download the latest version of the Gencove CLI."
            )

        if explorer_instances.results[0].status != "running":
            raise ValidationError(
                "Instance is not running. Cannot start shell session."
            )

        instance_id = str(explorer_instances.results[0].id)
        self.echo_debug("Requested explorer shell session credentials.")
        credentials = self.api_client.get_explorer_shell_session_credentials(
            instance_id=instance_id
        )

        self.start_ssm_session(credentials, instance_id)

    def start_ssm_session(
        self,
        credentials: ExplorerShellSessionCredentials,
        instance_id: str,
    ):
        """Start SSM session using AWS CLI.

        Args:
            credentials (ExplorerShellSessionCredentials): Credentials
                to send commands to the isntance.
            instance_id (str): Explorer instance id.

        Raises:
            ValidationError: If AWS CLI not available.
        """
        network_activity_background = Process(
            target=self.generate_network_activity,
            args=(credentials, instance_id),
        )
        network_activity_background.start()
        try:
            command = [
                "aws",
                "ssm",
                "start-session",
                "--target",
                credentials.ec2_instance_id,
                "--document-name",
                credentials.shell_session_ssm_document_name,
            ]
            with subprocess.Popen(  # nosec B603 (execution of untrusted input)
                command,
                env={
                    "AWS_ACCESS_KEY_ID": credentials.access_key,
                    "AWS_SECRET_ACCESS_KEY": credentials.secret_key,
                    "AWS_SESSION_TOKEN": credentials.token,
                    "AWS_DEFAULT_REGION": credentials.region_name,
                    "AWS_REGION": credentials.region_name,
                    "PATH": os.environ["PATH"],
                },
                stdin=sys.stdin,
                stdout=sys.stdout,
                stderr=sys.stderr,
            ) as process:

                def signal_handler(sig, frame):  # pylint: disable=unused-argument
                    if process.poll() is None:  # Check if the process is still running
                        process.send_signal(sig)

                signal.signal(signal.SIGINT, signal_handler)
                signal.signal(signal.SIGTERM, signal_handler)
                signal.signal(signal.SIGQUIT, signal_handler)
                signal.signal(signal.SIGHUP, signal_handler)

                process.wait()
        finally:
            if network_activity_background.is_alive():
                # Try to gracefully stop the backgroud process
                network_activity_background.terminate()

    def generate_network_activity(
        self,
        credentials: ExplorerShellSessionCredentials,
        instance_id: str,
    ):
        """Spoofs explorer instance network activity by sending random data.
        Runs forever.

        Args:
            credentials (ExplorerShellSessionCredentials): Credentials
                to send commands to the isntance.
            instance_id (str): Explorer instance id.
        """

        def ignore_signal(sig, frame):  # pylint: disable=unused-argument
            self.echo_debug(f"Ignoring SIGNAL in bg process: {sig}")

        signal.signal(signal.SIGINT, ignore_signal)

        def refresh_credentials():
            """Refresh credentials.

            Returns:
                ExplorerShellSessionCredentials: Session credentials.
            """
            return self.api_client.get_explorer_shell_session_credentials(
                instance_id=instance_id
            )

        boto3_session = get_boto_session_refreshable(refresh_credentials)
        ssm_client = boto3_session.client("ssm", region_name=credentials.region_name)
        while True:
            time.sleep(5)
            try:
                random_data = base64.b64encode(os.urandom(50_000)).decode()[:50_000]
                self.echo_debug(
                    f"Spoofing network activity random_data_len={len(random_data)}"
                )
                ssm_client.send_command(
                    InstanceIds=[credentials.ec2_instance_id],
                    DocumentName=credentials.network_activity_ssm_document_name,
                    Parameters={"RandomInputData": [random_data]},
                )
            except Exception as ex:  # pylint: disable=broad-except
                self.echo_debug(
                    f"Error while trying to spoof network activity: {repr(ex)}"
                )



"""Commands to be executed from command line."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import (
    Credentials,
    DOWNLOAD_TEMPLATE,
    DownloadTemplateParts,
)
from gencove.logger import echo_debug
from gencove.utils import enum_as_dict

from .constants import DownloadFilters, DownloadOptions
from .main import Download


@click.command(context_settings=dict(max_content_width=150))
@click.argument("destination")
@click.option("--project-id", help="Gencove project ID")
@click.option(
    "--sample-ids",
    help=(
        "A comma separated list of sample ids for "
        "which to download the deliverables. "
        "If not specified, download deliverables for all samples."
    ),
)
@click.option(
    "--file-types",
    help=(
        "A comma separated list of deliverable file types to download. "
        "If not specified, all file types will be downloaded."
    ),
)
@click.option(
    "--skip-existing/--no-skip-existing",
    default=True,
    help="Skip downloading files that already exist in DESTINATION",
)
@click.option(
    "--download-urls",
    help="Output a list of file urls available for download in a JSON format.",
    is_flag=True,
)
@click.option(
    "--download-template",
    default=DOWNLOAD_TEMPLATE,
    help=(
        "Change downloads structure. Defaults to: {}."
        "\nAvailable tokens: \n{}".format(
            DOWNLOAD_TEMPLATE,
            "\n".join(
                [f"{{{v}}}" for v in enum_as_dict(DownloadTemplateParts).values()]
            ),
        )
    ),
)
@add_options(common_options)
@click.option(
    "--no-progress",
    is_flag=True,
    help="If specified, no progress bar is shown.",
)
@click.option(
    "--checksums",
    is_flag=True,
    help="If specified, an additional checksum file will be downloaded for each deliverable.",  # noqa: E501 line too long pylint: disable=line-too-long
)
def download(  # pylint: disable=E0012,C0330,R0913
    destination,
    project_id,
    sample_ids,
    file_types,
    skip_existing,
    download_urls,
    download_template,
    host,
    email,
    password,
    api_key,
    no_progress,
    checksums,
):  # noqa: D413,D301,D412 # pylint: disable=C0301
    """Download deliverables of a project.

    `DESTINATION`: path/to/save/deliverables/to

    Must specify either project id or sample ids.

    Examples:

        Download all samples results:

            gencove download ./results --project-id d9eaa54b-aaac-4b85-92b0-0b564be6d7db

        Download some samples:

            gencove download ./results --sample-ids 59f5c1fd-cce0-4c4c-90e2-0b6c6c525d71,7edee497-12b5-4a1d-951f-34dc8dce1c1d

        Download specific deliverables:

            gencove download ./results --project-id d9eaa54b-aaac-4b85-92b0-0b564be6d7db --file-types alignment-bam,impute-vcf,fastq-r1,fastq-r2

        Skip download entirely and print out the deliverables as a JSON:

            gencove download - --project-id d9eaa54b-aaac-4b85-92b0-0b564be6d7db --download-urls
    """  # noqa: E501
    s_ids = tuple()
    if sample_ids:
        s_ids = tuple(s_id.strip() for s_id in sample_ids.split(","))
        echo_debug(f"Sample ids translation: {s_ids}")

    f_types = tuple()
    if file_types:
        f_types = tuple(f_type.strip() for f_type in file_types.split(","))
        echo_debug(f"File types translation: {f_types}")

    Download(
        destination,
        DownloadFilters(project_id=project_id, sample_ids=s_ids, file_types=f_types),
        Credentials(email=email, password=password, api_key=api_key),
        DownloadOptions(
            host=host,
            skip_existing=skip_existing,
            download_template=download_template,
        ),
        download_urls,
        no_progress,
        checksums,
    ).run()



"""VCR filters for projects tests."""
import json
from urllib.parse import urlparse

from gencove.tests.decorators import parse_response_to_json
from gencove.tests.filters import (
    _clean_query_params_from_filename,
    _replace_uuid_from_url,
)
from gencove.tests.utils import MOCK_UUID


@parse_response_to_json
def filter_list_projects_response(response, json_response):
    """Filter list project sensitive data from response."""
    if "results" in json_response:
        for result in json_response["results"]:
            if "id" in result:
                result["id"] = MOCK_UUID
            if "name" in result:
                result["name"] = "mock name"
            if "description" in result and result["description"]:
                result["description"] = "mock description"
            if "organization" in result:
                result["organization"] = MOCK_UUID
            if "sample_count" in result:
                result["sample_count"] = 1
            if "pipeline_capabilities" in result:
                result["pipeline_capabilities"] = MOCK_UUID
            if "roles" in result:
                result["roles"]["organization"]["id"] = MOCK_UUID
            if "webhook_url" in result and result["webhook_url"]:
                result["webhook_url"] = "http://mockurl.com"
    return response, json_response


def filter_pipeline_capabilities_request(request):
    """Filter pipeline capabilities sensitive data from request."""
    return _replace_uuid_from_url(request, "pipeline-capabilities")


@parse_response_to_json
def filter_pipeline_capabilities_response(response, json_response):
    """Filter pipeline capabilities sensitive data from response."""
    if "id" in json_response:
        json_response["id"] = MOCK_UUID
    if "name" in json_response:
        json_response["name"] = "mock name"
    return response, json_response


def filter_get_project_samples_request(request):
    """Filter project samples sensitive data from request."""
    return _replace_uuid_from_url(request, "project-samples")


@parse_response_to_json
def filter_get_project_samples_response(response, json_response):
    """Filter list project samples sensitive data from response."""
    if "results" in json_response:
        for result in json_response["results"]:
            if "id" in result:
                result["id"] = MOCK_UUID
            if "run" in result:
                result["run"] = MOCK_UUID
            if "client_id" in result:
                result["client_id"] = "mock client_id"
            if "physical_id" in result and result["physical_id"]:
                result["physical_id"] = "mock physical_id"
            if "legacy_id" in result and result["legacy_id"]:
                result["legacy_id"] = "mock legacy_id"
            if "last_status" in result:
                result["last_status"]["id"] = MOCK_UUID
                result["last_status"]["status"] = "mock status"
                if result["last_status"]["note"]:
                    result["last_status"]["note"] = "mock note"
            if "archive_last_status" in result:
                result["archive_last_status"]["id"] = MOCK_UUID
                result["archive_last_status"]["status"] = "mock status"
                if result["archive_last_status"]["transition_cutoff"]:
                    result["archive_last_status"][
                        "transition_cutoff"
                    ] = "mock transition_cutoff"
    return response, json_response


def filter_get_project_batch_types_request(request):
    """Filter project samples sensitive data from request."""
    return _replace_uuid_from_url(request, "project-batch-types")


def filter_post_project_restore_samples(request):
    """Filter project restore samples sensitive data from request."""
    if "project-restore-samples" in request.path:
        request = _replace_uuid_from_url(request, "project-restore-samples")
        try:
            body = json.loads(request.body)
            samples = [MOCK_UUID for _ in body["sample_ids"]]
            body["sample_ids"] = samples
            request.body = json.dumps(body).encode()
        except json.decoder.JSONDecodeError:
            pass
    return request


def filter_project_delete_samples(request):
    """Filter sensitive data from project-delete-samples request."""
    if "project-delete-samples" in request.path:
        request = _replace_uuid_from_url(request, "project-delete-samples")
        try:
            body = json.loads(request.body)
            samples = [MOCK_UUID for _ in body["sample_ids"]]
            body["sample_ids"] = samples
            request.body = json.dumps(body).encode()
        except json.decoder.JSONDecodeError:
            pass
    return request


def filter_projects_delete(request):
    """Filter sensitive data from projects-delete request."""
    if "projects-delete" in request.path:
        request = _replace_uuid_from_url(request, "projects-delete")
        try:
            body = json.loads(request.body)
            projects = [MOCK_UUID for _ in body["project_ids"]]
            body["project_ids"] = projects
            request.body = json.dumps(body).encode()
        except json.decoder.JSONDecodeError:
            pass
    return request


def filter_projects_detail_request(request):
    """Filter sensitive data from projects request."""
    return _replace_uuid_from_url(request, "projects")


@parse_response_to_json
def filter_projects_detail_response(response, json_response):
    """Filter project detail sensitive data from response."""
    if "id" in json_response:
        json_response["id"] = MOCK_UUID
    if "name" in json_response:
        json_response["name"] = "mock name"
    if "organization" in json_response:
        json_response["organization"] = MOCK_UUID
    if "pipeline_capabilities" in json_response:
        json_response["pipeline_capabilities"] = MOCK_UUID
    if "roles" in json_response and "organization" in json_response["roles"]:
        json_response["roles"]["organization"]["id"] = MOCK_UUID
    for file in json_response.get("files", []):
        if "id" in file:
            file["id"] = MOCK_UUID
        if "s3_path" in file:
            file["s3_path"] = f"mock/{file['s3_path'].split('/')[-1]}"
        if "size" in file and file["size"]:
            file["size"] = "1"
        if "download_url" in file:
            filename = _clean_query_params_from_filename(file["download_url"])
            file["download_url"] = f"https://example.com/{filename}"
    return response, json_response


def filter_project_batches_request(request):
    """Filter project batches sensitive data from request."""
    if "project-batches" in request.path:
        request = _replace_uuid_from_url(request, "project-batches")
        try:
            body = json.loads(request.body)
            body["name"] = "Mock name"
            body["batch_type"] = "Mock batch_type"
            body["sample_ids"] = [MOCK_UUID for _ in body["sample_ids"]]
            request.body = json.dumps(body).encode()
        except (json.decoder.JSONDecodeError, TypeError):
            pass
    return request


@parse_response_to_json
def filter_project_batches_response(response, json_response):
    """Filter project batches sensitive data from response."""
    if "results" in json_response:
        for result in json_response["results"]:
            if "id" in result:
                result["id"] = MOCK_UUID
            if "name" in result:
                result["name"] = "Mock name"
            if "batch_type" in result:
                result["batch_type"] = "Mock batch_type"
            if "last_status" in result:
                result["last_status"]["id"] = MOCK_UUID
            for file in result.get("files", []):
                if "id" in file:
                    file["id"] = MOCK_UUID
                if "s3_path" in file:
                    file["s3_path"] = "mock/file.zip"
                if "size" in file and file["size"]:
                    file["size"] = "1"
                if "download_url" in file:
                    file["download_url"] = "https://example.com/file.zip"
    return response, json_response


def filter_batches_request(request):
    """Filter batches sensitive data from request."""
    return _replace_uuid_from_url(request, "batches")


@parse_response_to_json
def filter_batches_response(response, json_response):
    """Filter batches sensitive data from response."""
    if "id" in json_response:
        json_response["id"] = MOCK_UUID
    if "name" in json_response:
        json_response["name"] = "mock name"
    if "batch_type" in json_response:
        json_response["batch_type"] = "Mock batch_type"
    if "sample_ids" in json_response:
        json_response["sample_ids"] = [MOCK_UUID for _ in json_response["sample_ids"]]
    if "last_status" in json_response:
        json_response["last_status"]["id"] = MOCK_UUID
    for file in json_response.get("files", []):
        if "id" in file:
            file["id"] = MOCK_UUID
        if "s3_path" in file:
            file["s3_path"] = f"mock/{file['s3_path'].split('/')[-1]}"
        if "size" in file and file["size"]:
            file["size"] = "1"
        if "download_url" in file:
            filename = urlparse(file["download_url"]).path.split("/")[-1]
            file["download_url"] = f"https://example.com/{filename}"
    return response, json_response


def filter_import_existing_samples_request(request):
    """Filter import_existing_samples sensitive data from request."""
    try:
        body = json.loads(request.body)
        if "project_id" in body:
            body["project_id"] = MOCK_UUID
        if "samples" in body:
            body["samples"] = [
                {"sample_id": MOCK_UUID, "client_id": "foo"} for _ in body["samples"]
            ]
        if "metadata" in body and body["metadata"]:
            body["metadata"] = {"foo": "bar"}
        request.body = json.dumps(body).encode()
    except (json.decoder.JSONDecodeError, TypeError):
        pass
    return request


@parse_response_to_json
def filter_import_existing_samples_response(response, json_response):
    """Filter import_existing_samples sensitive data from response."""
    if "project_id" in json_response:
        json_response["project_id"] = MOCK_UUID
    if "metadata" in json_response and json_response["metadata"]:
        json_response["metadata"] = {"foo": "bar"}
    if "samples" in json_response:
        for sample in json_response["samples"]:
            if "sample_id" in sample:
                sample["sample_id"] = MOCK_UUID
            if "client_id" in sample:
                sample["client_id"] = "foo"
    return response, json_response


@parse_response_to_json
def filter_project_pipelines_response(response, json_response):
    """Filter project_pipelines sensitive data from response."""
    if "results" in json_response:
        for result in json_response["results"]:
            if "id" in result:
                result["id"] = MOCK_UUID
            if "version" in result:
                result["version"] = "foo"
    if "meta" in json_response:
        if json_response["meta"]["next"]:
            next_url = urlparse(json_response["meta"]["next"])
            json_response["meta"][
                "next"
            ] = f"https://example.com{next_url.path}?{next_url.query}"
        if json_response["meta"]["previous"]:
            previous_url = urlparse(json_response["meta"]["previous"])
            json_response["meta"][
                "previous"
            ] = f"https://example.com{previous_url.path}?{previous_url.query}"
    return response, json_response


def filter_create_project_request(request):
    """Filter create_project sensitive data from request."""
    try:
        body = json.loads(request.body)
        if "pipeline_capabilities" in body:
            body["pipeline_capabilities"] = MOCK_UUID
        if "name" in body:
            body["name"] = "foo"
        request.body = json.dumps(body).encode()
    except (json.decoder.JSONDecodeError, TypeError):
        pass
    return request


@parse_response_to_json
def filter_create_project_response(response, json_response):
    """Filter create project sensitive data from response."""
    if "id" in json_response:
        json_response["id"] = MOCK_UUID
    if "name" in json_response:
        json_response["name"] = "mock name"
    if "description" in json_response and json_response["description"]:
        json_response["description"] = "mock description"
    if "organization" in json_response:
        json_response["organization"] = MOCK_UUID
    if "sample_count" in json_response:
        json_response["sample_count"] = 1
    if "pipeline_capabilities" in json_response:
        json_response["pipeline_capabilities"] = MOCK_UUID
    if "roles" in json_response:
        json_response["roles"]["organization"]["id"] = MOCK_UUID
    if "webhook_url" in json_response and json_response["webhook_url"]:
        json_response["webhook_url"] = "http://mockurl.com"
    return response, json_response


def filter_project_pipeline_capabilities_request(request):
    """Filter pipeline_capabilities sensitive data from request."""
    return _replace_uuid_from_url(request, "pipeline")


@parse_response_to_json
def filter_project_pipeline_capabilities_response(response, json_response):
    """Filter pipeline_capabilities sensitive data from response."""
    if "id" in json_response:
        json_response["id"] = MOCK_UUID
    if "version" in json_response:
        json_response["version"] = "mock version"
    if "capabilities" in json_response:
        for capability in json_response["capabilities"]:
            if "id" in capability:
                capability["id"] = MOCK_UUID
            if "name" in capability:
                capability["version"] = "mock name"
    return response, json_response


def filter_project_sample_manifest_request(request):
    """Filter project UUID data from request."""
    return _replace_uuid_from_url(request, "project-sample-manifests")


@parse_response_to_json
def filter_get_project_sample_manifests_response(response, json_response):
    """Filter pipeline_capabilities sensitive data from response."""
    for item in json_response:
        if "id" in item:
            item["id"] = MOCK_UUID
        if "s3_path" in item:
            item["version"] = "mock version"
        if "project" in item:
            item["project"] = MOCK_UUID
        if "file" in item:
            item["file"]["id"] = MOCK_UUID
            item["file"]["s3_path"] = "s3://cli-test/file.csv"
            item["file"]["download_url"] = "https://s3.amazonaws.com/file.csv"
    return response, json_response


@parse_response_to_json
def filter_get_sample_manifest_files_response(response, json_response):
    """Replace real filename with dummy data"""
    if "Content-Disposition" in response["headers"]:
        response["headers"]["Content-Disposition"] = "attachment; filename=foo.csv"
    return response, json_response


def filter_sample_manifests_request(request):
    """Filter sample manifests sensitive data from request."""
    if "amazonaws.com" in request.uri:
        request.uri = "https://v2-api-example.s3.amazonaws.com/example.csv"
    return request



"""Test project's batches list command."""
# pylint: disable=wrong-import-order, import-error
import io
import operator
import sys
from uuid import uuid4

from click import echo
from click.testing import CliRunner

from gencove.client import APIClient, APIClientTimeout  # noqa: I100
from gencove.command.projects.cli import list_project_batches
from gencove.models import ProjectBatches
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.projects.vcr.filters import (
    filter_project_batches_request,
    filter_project_batches_response,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_project_batches_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
            filter_project_batches_response,
        ],
    }


@pytest.mark.vcr
@assert_authorization
def test_list_project_batches__empty(credentials, mocker, project_id, recording, vcr):
    """Test project has not batches."""
    runner = CliRunner()
    if not recording:
        # Mock create_project_batch only if using the cassettes, since we
        # mock the return value.
        get_project_batches_response = get_vcr_response(
            "/api/v2/project-batches/", vcr, operator.contains
        )
        mocked_get_project_batches = mocker.patch.object(
            APIClient,
            "get_project_batches",
            return_value=ProjectBatches(**get_project_batches_response),
        )
    res = runner.invoke(
        list_project_batches,
        [project_id, *credentials],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_get_project_batches.assert_called_once()
    assert res.output == ""


@pytest.mark.vcr
@assert_authorization
def test_list_project_batches__not_empty(
    credentials, mocker, project_id_batches, recording, vcr
):
    """Test project batches being outputed to the shell."""
    runner = CliRunner()
    if not recording:
        # Mock create_project_batch only if using the cassettes, since we
        # mock the return value.
        get_project_batches_response = get_vcr_response(
            "/api/v2/project-batches/", vcr, operator.contains
        )
        mocked_get_project_batches = mocker.patch.object(
            APIClient,
            "get_project_batches",
            return_value=ProjectBatches(**get_project_batches_response),
        )
    res = runner.invoke(
        list_project_batches,
        [project_id_batches, *credentials],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_get_project_batches.assert_called_once()

        output_line = io.BytesIO()
        sys.stdout = output_line
        for response in ProjectBatches(**get_project_batches_response).results:
            echo(
                "\t".join(
                    [
                        str(response.id),
                        response.last_status.created.isoformat(),
                        response.last_status.status,
                        response.batch_type,
                        response.name,
                    ]
                )
            )
        assert output_line.getvalue() == res.output.encode()


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_project_batches__not_empty_slow_response_retry(credentials, mocker):
    """Test project batches slow response retry."""
    runner = CliRunner()
    mocked_get_project_batches = mocker.patch.object(
        APIClient,
        "get_project_batches",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )

    res = runner.invoke(
        list_project_batches,
        [str(uuid4()), *credentials],
    )
    assert res.exit_code == 1
    assert mocked_get_project_batches.call_count == 2



"""Configure inactivity stop for explorer instances command definition.
"""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import StopInstanceInactivity


@click.command("inactivity-stop")
@click.option(
    "--hours",
    default=None,
    help=(
        "Amount of hours of inactivity to wait before stopping instances. "
        "If hours == 0 then the instances are not stopped. "
        "If hours == None then the org-level config is applied."
    ),
)
@click.option(
    "--organization",
    is_flag=True,
    help="Sets the config for the entire Organization.",
)
@click.option(
    "--override",
    default=None,
    type=bool,
    help="Organization config overrides the instances configs.",
)
@add_options(common_options)
def inactivity_stop(  # pylint: disable=too-many-arguments
    hours: str,
    organization: bool,
    override: bool,
    host,
    email,
    password,
    api_key,
):  # pylint: disable=line-too-long
    """Stop instances after hours of inactivity.
    An instance can be configured to be stopped after some time of inactivity.
    Aditionally, this configuration can also be set at an Organization-level.
    By default the instance configuration has precedence over the org-level config,
    but this can be changed with the --override flag.
    """  # noqa: E501
    StopInstanceInactivity(
        hours,
        organization,
        override,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()
